# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import importlib.util
import pathlib
import re


class OptionTemplate:
    """
    Template to be used by derived options - default and custom ones
    """
    number = 0

    def __init__(self, number):
        self.number = number

    def build(self, **kwargs):
        """
        to be filled with life by every single option
        every option has its special treatment of input and output data per request
        return default dummy values
        """
        return '', False

    def initialize(self, **kwargs):
        """
        to be filled with life by every single option
        every transaction has the opportunity to add options, depending on request
        """
        pass

    @staticmethod
    def convert_to_string(number, payload):
        """
        glue option number with payload
        """
        # option number and length take 2 byte each so the string has to be 4 chars long
        option_string = f'{number:04x}'  # option number
        option_string += f'{(len(payload)//2):04x}'  # payload length, /2 because 2 chars are 1 byte
        option_string += payload
        return option_string


# globally available options
OPTIONS = {}
options_path = pathlib.Path(__file__).parent
pattern = re.compile('option_[0-9]{1,3}$')

# get all option files in path and put them into options dict
for path in options_path.glob('option_*.py'):
    # get rid of ".py" because this suffix won't be in option dict anyway
    name = path.name.rstrip(path.suffix)
    if re.match(pattern, name):
        # load option module
        spec = importlib.util.spec_from_file_location(name, path)
        option = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(option)
        number = int(name.split('_')[1])
        # add to global options constant
        OPTIONS[number] = option.Option(number)



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

from ..config import (Address,
                      cfg,
                      Prefix)
from ..constants import CONST
from ..domain import get_ip_from_dns

from .parse_pattern import (parse_pattern_address,
                            parse_pattern_prefix)


def from_config(client=None, client_config=None, transaction=None):
    # give client hostname + class
    client.hostname = client_config.HOSTNAME
    client.client_class = client_config.CLASS
    # apply answer type of client to transaction - useful if no answer or no address available is configured
    transaction.answer = cfg.CLASSES[client.client_class].ANSWER
    # continue only if request interface matches class interfaces
    if transaction.interface in cfg.CLASSES[client.client_class].INTERFACE:
        # if fixed addresses are given build them
        if client_config.ADDRESS is not None and \
                    CONST.OPTION.IA_NA in transaction.ia_options:
            for address in client_config.ADDRESS:
                if len(address) > 0:
                    # fixed addresses are assumed to be non-temporary
                    #
                    # todo: lifetime of address should be set by config too
                    #
                    ia = Address(address=address,
                                 ia_type='na',
                                 preferred_lifetime=cfg.PREFERRED_LIFETIME,
                                 valid_lifetime=cfg.VALID_LIFETIME,
                                 category='fixed',
                                 aclass='fixed',
                                 atype='fixed')
                    client.addresses.append(ia)

        # if fixed prefixes are given and requested build them
        if client_config.PREFIX is not None and \
                    CONST.OPTION.IA_PD in transaction.ia_options:
            for prefix in client_config.PREFIX:
                ia_pd = Prefix(prefix=prefix['address'],
                               length=prefix['length'],
                               preferred_lifetime=cfg.PREFERRED_LIFETIME,
                               valid_lifetime=cfg.VALID_LIFETIME,
                               route_link_local=False)
                client.prefixes.append(ia_pd)

        if not client_config.CLASS == '':
            # add all addresses which belong to that class
            for address in cfg.CLASSES[client_config.CLASS].ADDRESSES:
                # addresses of category 'dns' will be searched in DNS
                if cfg.ADDRESSES[address].CATEGORY == 'dns':
                    a = get_ip_from_dns(client.hostname)
                else:
                    a = parse_pattern_address(cfg.ADDRESSES[address], client_config, transaction)
                # in case range has been exceeded a will be None
                if a:
                    ia = Address(address=a,
                                 ia_type=cfg.ADDRESSES[address].IA_TYPE,
                                 preferred_lifetime=cfg.ADDRESSES[address].PREFERRED_LIFETIME,
                                 valid_lifetime=cfg.ADDRESSES[address].VALID_LIFETIME,
                                 category=cfg.ADDRESSES[address].CATEGORY,
                                 aclass=cfg.ADDRESSES[address].CLASS,
                                 atype=cfg.ADDRESSES[address].TYPE,
                                 dns_update=cfg.ADDRESSES[address].DNS_UPDATE,
                                 dns_zone=cfg.ADDRESSES[address].DNS_ZONE,
                                 dns_rev_zone=cfg.ADDRESSES[address].DNS_REV_ZONE,
                                 dns_ttl=cfg.ADDRESSES[address].DNS_TTL)
                    client.addresses.append(ia)

            # add all bootfiles which belong to that class
            for bootfile in cfg.CLASSES[client_config.CLASS].BOOTFILES:
                client_architecture = cfg.BOOTFILES[bootfile].CLIENT_ARCHITECTURE
                user_class = cfg.BOOTFILES[bootfile].USER_CLASS

                # check if transaction attributes matches the bootfile defintion
                if (not client_architecture or
                    transaction.client_architecture == client_architecture or
                    transaction.known_client_architecture == client_architecture) and \
                        (not user_class or
                         transaction.user_class == user_class):
                    client.bootfiles.append(cfg.BOOTFILES[bootfile])

            # if prefixes are advertised in this class and the client demands a prefix (trough IA_PD)
            if 'prefixes' in cfg.CLASSES[client_config.CLASS].ADVERTISE and \
                    CONST.OPTION.IA_PD in transaction.ia_options:
                for prefix in cfg.CLASSES[client_config.CLASS].PREFIXES:
                    p = parse_pattern_prefix(cfg.PREFIXES[prefix], client_config, transaction)
                    # in case range has been exceeded p will be None
                    if p:
                        ia_pd = Prefix(prefix=p,
                                       length=cfg.PREFIXES[prefix].LENGTH,
                                       preferred_lifetime=cfg.PREFIXES[prefix].PREFERRED_LIFETIME,
                                       valid_lifetime=cfg.PREFIXES[prefix].VALID_LIFETIME,
                                       category=cfg.PREFIXES[prefix].CATEGORY,
                                       pclass=cfg.PREFIXES[prefix].CLASS,
                                       ptype=cfg.PREFIXES[prefix].TYPE,
                                       route_link_local=cfg.PREFIXES[prefix].ROUTE_LINK_LOCAL)
                        client.prefixes.append(ia_pd)

        if client_config.ADDRESS == client_config.CLASS == '':
            # use default class if no class or address is given
            for address in cfg.CLASSES['default_' + transaction.interface].ADDRESSES:
                client.client_class = 'default_' + transaction.interface
                # addresses of category 'dns' will be searched in DNS
                if cfg.ADDRESSES[address].CATEGORY == 'dns':
                    a = get_ip_from_dns(client.hostname)
                else:
                    a = parse_pattern_address(cfg.ADDRESSES[address], client_config, transaction)
                if a:
                    ia = Address(address=a, ia_type=cfg.ADDRESSES[address].IA_TYPE,
                                 preferred_lifetime=cfg.ADDRESSES[address].PREFERRED_LIFETIME,
                                 valid_lifetime=cfg.ADDRESSES[address].VALID_LIFETIME,
                                 category=cfg.ADDRESSES[address].CATEGORY,
                                 aclass=cfg.ADDRESSES[address].CLASS,
                                 atype=cfg.ADDRESSES[address].TYPE,
                                 dns_update=cfg.ADDRESSES[address].DNS_UPDATE,
                                 dns_zone=cfg.ADDRESSES[address].DNS_ZONE,
                                 dns_rev_zone=cfg.ADDRESSES[address].DNS_REV_ZONE,
                                 dns_ttl=cfg.ADDRESSES[address].DNS_TTL)
                    client.addresses.append(ia)

            for bootfile in cfg.CLASSES['default_' + transaction.interface].BOOTFILES:
                client_architecture = bootfile.CLIENT_ARCHITECTURE
                user_class = bootfile.USER_CLASS

                # check if transaction attributes match the bootfile definition
                if (not client_architecture or
                    transaction.client_architecture == client_architecture or
                    transaction.known_client_architecture == client_architecture) and \
                        (not user_class or
                         transaction.user_class == user_class):
                    client.bootfiles.append(bootfile)

    # given client has been modified successfully
    return True



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import subprocess
import sys
from threading import Thread
import time
import traceback

import dns.query
import dns.resolver
import dns.reversename
import dns.update

from .config import cfg
from .globals import (collected_macs,
                      dns_query_queue,
                      keyring,
                      requests,
                      requests_blacklist,
                      route_queue,
                      resolver_update,
                      timer,
                      transactions)
from .helpers import colonify_ip6
from .log import log
from .storage import volatile_store


class DNSQueryThread(Thread):
    """
        thread for updating DNS entries of valid leases without blocking main thread
    """

    def __init__(self):
        Thread.__init__(self, name='DNSQuery')
        self.setDaemon(True)

    def run(self):
        # wait for new queries in queue until the end of the world
        while True:
            action, hostname, a = dns_query_queue.get()
            # colonify address for DNS
            address = colonify_ip6(a.ADDRESS)
            try:
                # update AAAA record, delete old entry first
                update = dns.update.Update(a.DNS_ZONE, keyring=keyring)
                update.delete(hostname, 'AAAA')
                # if DNS should be updated do it - not the case if IP is released
                if action == 'update':
                    update.add(hostname, a.DNS_TTL, 'AAAA', address)
                dns.query.tcp(update, cfg.DNS_UPDATE_NAMESERVER)

                # the reverse record will be first checked if it points
                # to the current hostname, if not, it will be deleted first
                update_rev = dns.update.Update(a.DNS_REV_ZONE, keyring=keyring)
                try:
                    answer = resolver_update.query(dns.reversename.from_address(address), 'PTR')
                    for rdata in answer:
                        hostname_ns = str(rdata).split('.')[0]
                        # if ip address is related to another host delete this one
                        if hostname_ns != hostname:
                            update_rev.delete(dns.reversename.from_address(address), 'PTR',
                                              hostname_ns + '.' + a.DNS_ZONE + '.')
                except dns.resolver.NXDOMAIN:
                    log.error(f'Received NXDOMAIN when trying to resolve {address}')
                # if DNS should be updated do it - not the case if IP is released
                if action == 'update':
                    update_rev.add(dns.reversename.from_address(address), a.DNS_TTL, 'PTR',
                                   hostname + '.' + a.DNS_ZONE + '.')
                elif action == 'release':
                    update_rev.delete(dns.reversename.from_address(address), 'PTR')
                dns.query.tcp(update_rev, cfg.DNS_UPDATE_NAMESERVER)
            except Exception as err:
                traceback.print_exc(file=sys.stdout)
                sys.stdout.flush()
                log.error('DNSUPDATE: ' + str(err))


class TidyUpThread(Thread):
    """
        clean leases and transactions if obsolete
    """

    def __init__(self):
        Thread.__init__(self, name='TidyUp')
        self.setDaemon(True)

    def run(self):
        try:
            # counter for database cleaning interval
            dbcount = 0

            # get and delete invalid leases
            while True:
                # transaction data can be deleted after transaction is finished
                for transaction in list(transactions.values()):
                    try:
                        if timer.time > transaction.timestamp + cfg.CLEANING_INTERVAL * 10:
                            transactions.pop(transaction.client_llip + transaction.id)
                    except Exception as err:
                        log.error(f'TidyUp: transaction {str(err)} has already been deleted')
                        traceback.print_exc(file=sys.stdout)
                        sys.stdout.flush()

                # if disconnected try reconnect
                if not volatile_store.connected:
                    volatile_store.db_connect()
                else:
                    # cleaning database once per minute should be enough
                    if dbcount > 60 // cfg.CLEANING_INTERVAL:
                        # remove leases which might not be recycled like random addresses for example
                        volatile_store.remove_leases(timer.time, 'random')
                        # set leases and prefixes free whose valid lifetime is over
                        volatile_store.release_free_leases(timer.time)
                        volatile_store.release_free_prefixes(timer.time)
                        # unlock advertised leases and prefixes remaining
                        volatile_store.unlock_unused_advertised_leases(timer.time)
                        volatile_store.unlock_unused_advertised_prefixes(timer.time)
                        # remove routes with inactive prefixes
                        self.check_routes()
                        # check for brute force clients and put them into blacklist if necessary
                        self.check_requests(timer.time)
                        dbcount = 0
                dbcount += 1

                # clean collected MAC addresses after 300 seconds
                # some Linuxes seem to be pretty slow and run out of the previous 30 seconds
                if not cfg.CACHE_MAC_LLIP:
                    timestamp = timer.time
                    for record in list(collected_macs.values()):
                        if record.timestamp + 60 * cfg.CLEANING_INTERVAL < timestamp:
                            if cfg.LOG_MAC_LLIP:
                                log.info(f'deleted mac {record.mac} for llip {colonify_ip6(record.llip)}')
                            collected_macs.pop(record.llip)
                    del timestamp
                time.sleep(cfg.CLEANING_INTERVAL)
        except:
            traceback.print_exc(file=sys.stdout)
            sys.stdout.flush()

    @staticmethod
    def check_routes():
        """
            remove routes with inactive prefixes
            thanks to PyCharm this might be a @staticmethod
        """
        for prefix in volatile_store.get_inactive_prefixes():
            length, router, pclass = volatile_store.get_route(prefix)
            # hopefully the class stored in database still exists
            if pclass in cfg.CLASSES:
                route_queue.put(('down', cfg.CLASSES[pclass].CALL_DOWN, prefix, length, router))

    @staticmethod
    def check_requests(now):
        """
            check for brute force clients and put them into blacklist if necessary
            get time as now from caller
            dito here regarding MAC addresses
        """
        # clean blacklist
        for client in list(requests_blacklist.keys()):
            if now > requests_blacklist[client].timestamp + cfg.REQUEST_LIMIT_RELEASE_TIME:
                log.info(f"Releasing client {client} from blacklist")
                requests_blacklist.pop(client)

        # clean default requests list
        for client in list(requests.keys()):
            if now > requests[client].timestamp + cfg.REQUEST_LIMIT_TIME:
                if requests[client].count > cfg.REQUEST_LIMIT_COUNT:
                    log.info(f"Blacklisting client {client} after {requests[client].count} requests")
                    requests_blacklist[client] = requests.pop(client)
                else:
                    requests.pop(client)


class RouteThread(Thread):
    """
        thread for updating routes without blocking main thread
    """

    def __init__(self, route_queue):
        Thread.__init__(self, name='Route')
        self.setDaemon(True)
        self.route_queue = route_queue

    def run(self):
        """
            wait for new queries in queue until the end of the world
        """
        while True:
            mode, call, prefix, length, router = self.route_queue.get()
            call_real = call.replace('$prefix$', colonify_ip6(prefix)). \
                replace('$length$', str(length)). \
                replace('$router$', colonify_ip6(router)). \
                replace('$mode$', mode)
            # subprocess needs list as argument which it gets by split()
            try:
                result = subprocess.call(call_real.split(' '))
            except Exception as err:
                result = err
            # ignore result to avoid routes being set but not noted in database when a command like
            # 'ip -6 route delete' gives a return code not 0 because a route already exists
            if mode == 'up':
                volatile_store.store_route(prefix, length, router, timer.time)
            if mode == 'down':
                volatile_store.remove_route(prefix)
            log.info(f"Called '{call_real}' to modify route - result: {result}")


class TimerThread(Thread):
    """
        thread for timer, used in different places
    """

    def __init__(self):
        Thread.__init__(self, name='Timer')
        self.setDaemon(True)

    def run(self):
        while True:
            # set globally available time here to new value
            timer.time = time.time()
            time.sleep(1)



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

from binascii import hexlify
from socket import (AF_INET6,
                    inet_pton)

from dhcpy6d import collected_macs
from dhcpy6d.client import Client
from dhcpy6d.config import cfg
from dhcpy6d.constants import CONST
from dhcpy6d.helpers import (colonify_ip6,
                             combine_prefix_length)
from dhcpy6d.options import OptionTemplate


class Option(OptionTemplate):
    """
    Option 25 Prefix Delegation
    """
    def build(self, transaction=None, **kwargs):
        # dummy empty defaults
        response_string_part = ''
        options_answer_part = None

        # check if MAC of LLIP is really known
        if transaction.client_llip in collected_macs or cfg.IGNORE_MAC:
            # collect client information
            if transaction.client is None:
                transaction.client = Client(transaction)

            # Only if prefixes are provided
            if 'prefixes' in cfg.CLASSES[transaction.client.client_class].ADVERTISE:
                # check if only a short NoPrefixAvail answer or none at all is to be returned
                if not transaction.answer == 'normal':
                    if transaction.answer == 'noprefix':
                        # Option 13 Status Code Option - statuscode is 6: 'No Prefix available'
                        response_string_part = self.convert_to_string(CONST.OPTION.STATUS_CODE,
                                                                      f'{CONST.STATUS.NO_PREFIX_AVAILABLE:04x}')
                        # clean client prefixes which not be deployed anyway
                        transaction.client.prefixes[:] = []
                        # options in answer to be logged
                        options_answer_part = self.number
                else:
                    # if client could not be built because of database problems send
                    # status message back
                    if transaction.client:
                        # embed option 26 into option 25 - several if necessary
                        ia_prefixes = ''
                        try:
                            for prefix in transaction.client.prefixes:
                                ipv6_prefix = hexlify(inet_pton(AF_INET6, colonify_ip6(prefix.PREFIX))).decode()
                                if prefix.VALID:
                                    preferred_lifetime = f'{int(prefix.PREFERRED_LIFETIME):08x}'
                                    valid_lifetime = f'{int(prefix.VALID_LIFETIME):08x}'
                                else:
                                    preferred_lifetime = f'{0:08x}'
                                    valid_lifetime = f'{0:08x}'
                                length = f'{int(prefix.LENGTH):02x}'
                                ia_prefixes += self.convert_to_string(CONST.OPTION.IAPREFIX,
                                                                      preferred_lifetime +
                                                                      valid_lifetime +
                                                                      length +
                                                                      ipv6_prefix)

                            if transaction.client.client_class != '':
                                t1 = f'{int(cfg.CLASSES[transaction.client.client_class].T1):08x}'
                                t2 = f'{int(cfg.CLASSES[transaction.client.client_class].T2):08x}'
                            else:
                                t1 = f'{int(cfg.T1):08x}'
                                t2 = f'{int(cfg.T2):08x}'

                            # even if there are no prefixes server has to deliver an empty PD
                            response_string_part = self.convert_to_string(self.number,
                                                                          transaction.iaid +
                                                                          t1 +
                                                                          t2 +
                                                                          ia_prefixes)
                            # if no prefixes available a NoPrefixAvail status code has to be sent
                            if ia_prefixes == '':
                                # REBIND not possible
                                if transaction.last_message_received_type == CONST.MESSAGE.REBIND:
                                    # Option 13 Status Code Option - statuscode is 3: 'NoBinding'
                                    response_string_part += self.convert_to_string(CONST.OPTION.STATUS_CODE,
                                                                                   f'{CONST.STATUS.NO_BINDING:04x}')
                                else:
                                    # Option 13 Status Code Option - statuscode is 6: 'No Prefix available'
                                    response_string_part += self.convert_to_string(  # break because line too long
                                                                              CONST.OPTION.STATUS_CODE,
                                                                              f'{CONST.STATUS.NO_PREFIX_AVAILABLE:04x}')
                            # options in answer to be logged
                            options_answer_part = self.number

                        except Exception as err:
                            print(err)
                            # Option 13 Status Code Option - statuscode is 6: 'No Prefix available'
                            response_string_part = self.convert_to_string(CONST.OPTION.STATUS_CODE,
                                                                          f'{CONST.STATUS.NO_PREFIX_AVAILABLE:04x}')
                            # options in answer to be logged
                            options_answer_part = self.number
                    else:
                        # Option 13 Status Code Option - statuscode is 6: 'No Prefix available'
                        response_string_part = self.convert_to_string(CONST.OPTION.STATUS_CODE,
                                                                      f'{CONST.STATUS.NO_PREFIX_AVAILABLE:04x}')
                        # options in answer to be logged
                        options_answer_part = self.number

        return response_string_part, options_answer_part

    def initialize(self, transaction=None, option=None, **kwargs):
        for payload in option:
            # iaid        t1        t2       ia_prefix   opt_length       preferred validlt    length    prefix
            # 00000001    ffffffff  ffffffff  001a        0019             00000e10   00001518    30     fd66123400....
            # 8               16      24      28          32                  40      48          50      82
            transaction.iaid = payload[0:8]
            transaction.iat1 = int(payload[8:16], 16)
            transaction.iat2 = int(payload[16:24], 16)
            # Prefixes given by client if any
            for p in range(len(payload[32:])//50):
                prefix = payload[50:][(p*58):(p*58)+32]
                length = int(payload[48:][(p*58):(p*58)+2], 16)
                prefix_combined = combine_prefix_length(prefix, length)
                # in case a prefix is asked for twice by one host ignore the twin
                if prefix_combined not in transaction.prefixes:
                    transaction.prefixes.append(prefix_combined)
                del(prefix, length, prefix_combined)
        transaction.ia_options.append(CONST.OPTION.IA_PD)



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

from dhcpy6d.constants import CONST
from dhcpy6d.options import OptionTemplate


class Option(OptionTemplate):
    """
    61 Client System Architecture Type
    """
    def initialize(self, transaction=None, option=None, **kwargs):
        # raw client architecture is supplied as a 16-bit integer (e. g. 0007)
        # See https://tools.ietf.org/html/rfc4578#section-2.1
        transaction.client_architecture = option
        # short number (0007 => 7 for dictionary usage)
        client_architecture_short = int(transaction.client_architecture)
        if client_architecture_short in CONST.ARCHITECTURE_TYPE_DICT:
            transaction.known_client_architecture = CONST.ARCHITECTURE_TYPE_DICT[client_architecture_short]



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import queue
import platform
import time

import dns.resolver
import dns.tsigkeyring

from .config import cfg
from .constants import CONST

# dummy value is None for DNS related
resolver_query = None
resolver_update = None
keyring = None

# if nameserver is given create resolver
if len(cfg.NAMESERVER) > 0:
    # default nameservers for DNS queries
    resolver_query = dns.resolver.Resolver()
    resolver_query.nameservers = cfg.NAMESERVER

# RNDC Key for DNS updates from ISC Bind /etc/rndc.key
if cfg.DNS_UPDATE:
    if cfg.DNS_USE_RNDC:
        keyring = dns.tsigkeyring.from_text({cfg.DNS_RNDC_KEY: cfg.DNS_RNDC_SECRET})
    # resolver for DNS updates
    resolver_update = dns.resolver.Resolver()
    resolver_update.nameservers = [cfg.DNS_UPDATE_NAMESERVER]


class Timer:
    """
    global object containing time set by TimerThread
    """
    __time = 0

    def __init__(self):
        self.time = time.time()

    @property
    def time(self):
        return self.__time

    @time.setter
    def time(self, new_time):
        self.__time = int(new_time)


# global time variable, synchronized by TimerThread
timer = Timer()

# dictionary to store transactions - key is transaction ID, value a transaction object
transactions = {}
# collected MAC addresses from clients, mapping to link local IPs
collected_macs = {}

# queues for queries
config_query_queue = queue.Queue()
config_answer_queue = queue.Queue()
volatile_query_queue = queue.Queue()
volatile_answer_queue = queue.Queue()

# queue for dns actualization
dns_query_queue = queue.Queue()

# queue for executing some script to modify routes after delegating prefixes
route_queue = queue.Queue()

# attempt to log connections and count them to find out which clients do silly crazy brute force
requests = {}
requests_blacklist = {}

# save OS
OS = platform.system()
if 'BSD' in OS:
    OS = 'BSD'

# platform-dependant neighbor cache call
# every platform has its different output
# dev, llip and mac are positions of output of call
# len is minimal length a line has to have to be evaluable
#
# update: has been different to Linux which now access neighbor cache natively
NC = {'BSD': {'call': '/usr/sbin/ndp -a -n',
              'dev': 2,
              'llip': 0,
              'mac': 1,
              'len': 3},
      'Darwin': {'call': '/usr/sbin/ndp -a -n',
                 'dev': 2,
                 'llip': 0,
                 'mac': 1,
                 'len': 3}
      }

# libc access via ctypes, needed for interface handling, get it by helpers.get_libc()
# obsolete in Python 3
# LIBC = get_libc()

# index IF name > number, gets filled in UDPMulticastIPv6
IF_NAME = {}
# index IF number > name
IF_NUMBER = {}

# IA_NA, IA_TA and IA_PD Options referred here in handler
IA_OPTIONS = (CONST.OPTION.IA_NA,
              CONST.OPTION.IA_TA,
              CONST.OPTION.IA_PD)

# options to be ignored when logging
IGNORED_LOG_OPTIONS = ['options_raw', 'client', 'client_config_dicts', 'timestamp', 'iat1', 'iat2', 'id']

# empty options string test
EMPTY_OPTIONS = [None, False, '', []]

# dummy IAID for transactions
DUMMY_IAID = '00000000'

# dummy MAC for transactions
DUMMY_MAC = '00:00:00:00:00:00'

# store
# because of thread trouble there should not be too much db connections at once
# so we need to use the queryqueue way - subject to change
# source of configuration of hosts
# use client configuration only if needed
config_store = volatile_store = None



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import sys
import traceback

from ..config import cfg
from ..helpers import error_exit

from .schemas import POSTGRESQL_SCHEMA
from .store import DB


class DBPostgreSQL(DB):
    """
    PostgreSQL connection - to be tested!
    """
    # different to default derived MYSQL_SQLITE schema
    SCHEMAS = POSTGRESQL_SCHEMA
    QUERY_TABLES = f"SELECT table_name FROM information_schema.tables WHERE " \
                   f"table_schema = 'public' AND " \
                   f"table_catalog = '{cfg.STORE_DB_DB}'"

    def db_connect(self):
        """
            Connect to database server according to database type
        """
        try:
            if 'psycopg2' not in sys.modules:
                import psycopg2
            self.db_module = sys.modules['psycopg2']
        except:
            traceback.print_exc(file=sys.stdout)
            sys.stdout.flush()
            error_exit('ERROR: Cannot find module psycopg2. Please install to proceed.')
        try:
            self.connection = self.db_module.connect(host=cfg.STORE_DB_HOST,
                                                     database=cfg.STORE_DB_DB,
                                                     user=cfg.STORE_DB_USER,
                                                     password=cfg.STORE_DB_PASSWORD)
            self.connection.autocommit = True
            self.cursor = self.connection.cursor()
            self.connected = True
        except:
            traceback.print_exc(file=sys.stdout)
            sys.stdout.flush()
            self.connected = False
        return self.connected

    def db_query(self, query):
        """
        execute DB query
        """
        try:
            self.cursor.execute(query)
        # catch impossible INSERTs
        except self.db_module.errors.UniqueViolation:
            return 'INSERT_ERROR'
        except self.db_module.errors.IntegrityError:
            return 'INSERT_ERROR'
        except Exception as err:
            # try to reestablish database connection
            print(f'Error: {str(err)}')
            if not self.db_connect():
                return None
            else:
                try:
                    self.cursor.execute(query)
                except:
                    traceback.print_exc(file=sys.stdout)
                    sys.stdout.flush()
                    self.connected = False
                    return None
        try:
            result = self.cursor.fetchall()
        # If there is no result after a database reconnect a None would lead to eternal loop
        except self.db_module.ProgrammingError:
            return []
        except Exception as err:
            return None
        return result



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import configparser

from ..config import cfg
from ..helpers import (decompress_ip6,
                       error_exit,
                       listify_option, convert_prefix_inline)

from .store import (ClientConfig,
                    Store)


class Textfile(Store):
    """
        client config in text files
    """
    def __init__(self, query_queue, answer_queue):
        Store.__init__(self, query_queue, answer_queue)
        self.connection = None

        # store config information of hosts
        self.hosts = {}
        self.index_mac = {}
        self.index_duid = {}

        # store IDs for ID-based hosts to check if there are duplicates
        self.ids = {}

        # instantiate a Configparser
        config = configparser.ConfigParser()
        config.read(cfg.STORE_FILE_CONFIG)

        # read all sections of config file
        # a section here is a host
        for section in config.sections():
            hostname = config[section]['hostname'].lower()
            # only if section matches hostname the following steps are of any use
            if section.lower() == hostname:
                self.hosts[hostname] = ClientConfig()
                for item in config.items(hostname):
                    # lowercase all MAC addresses, DUIDs, IPv6 addresses and prefixes
                    if item[0].upper() in ['ADDRESS', 'DUID', 'HOSTNAME', 'MAC', 'PREFIX']:
                        self.hosts[hostname].__setattr__(item[0].upper(), str(item[1]).lower())
                    else:
                        self.hosts[hostname].__setattr__(item[0].upper(), str(item[1]))

                # Test if host has ID
                if self.hosts[hostname].CLASS in cfg.CLASSES:
                    for a in cfg.CLASSES[self.hosts[hostname].CLASS].ADDRESSES:
                        if cfg.ADDRESSES[a].CATEGORY == 'id' and self.hosts[hostname].ID == '':
                            error_exit(f"Textfile client configuration: No ID given "
                                       f"for client '{self.hosts[hostname].HOSTNAME}'")
                else:
                    error_exit(f"Textfile client configuration: Class '{self.hosts[hostname].CLASS}' "
                               f"of host '{self.hosts[hostname].HOSTNAME}' is not defined")

                if self.hosts[hostname].ID != '':
                    if self.hosts[hostname].ID in list(self.ids.keys()):
                        error_exit(f"Textfile client configuration: ID '{self.hosts[hostname].ID}' "
                                   f"of client '{self.hosts[hostname].HOSTNAME}' is already used "
                                   f"by '{self.ids[self.hosts[hostname].ID]}'.")
                    else:
                        self.ids[self.hosts[hostname].ID] = self.hosts[hostname].HOSTNAME

                # in case of various MAC addresses split them...
                self.hosts[hostname].MAC = listify_option(self.hosts[hostname].MAC)

                # in case of various fixed addresses split them and avoid decompressing of ':'...
                self.hosts[hostname].ADDRESS = listify_option(self.hosts[hostname].ADDRESS)

                # Decompress IPv6-Addresses
                if self.hosts[hostname].ADDRESS is not None:
                    self.hosts[hostname].ADDRESS = [decompress_ip6(x) for x in self.hosts[hostname].ADDRESS]

                # in case of multiple supplied prefixes convert them to list
                self.hosts[hostname].PREFIX = listify_option(self.hosts[hostname].PREFIX)

                # split prefix into address and length, verify address
                if self.hosts[hostname].PREFIX is not None:
                    self.hosts[hostname].PREFIX = [convert_prefix_inline(x) for x in self.hosts[hostname].PREFIX]

                # and put the host objects into index
                if self.hosts[hostname].MAC:
                    for m in self.hosts[hostname].MAC:
                        if m not in self.index_mac:
                            self.index_mac[m] = [self.hosts[hostname]]
                        else:
                            self.index_mac[m].append(self.hosts[hostname])

                # add DUIDs to IndexDUID
                if not self.hosts[hostname].DUID == '':
                    if not self.hosts[hostname].DUID in self.index_duid:
                        self.index_duid[self.hosts[hostname].DUID] = [self.hosts[hostname]]
                    else:
                        self.index_duid[self.hosts[hostname].DUID].append(self.hosts[hostname])
            else:
                error_exit(f"Textfile client configuration: section [{section.lower()}] "
                           f"does not match hostname '{hostname}'")
        # not very meaningful in case of databaseless textfile config but for completeness
        self.connected = True

    def get_client_config_by_mac(self, transaction):
        """
            get host(s?) and its information belonging to that mac
        """
        hosts = list()
        mac = transaction.mac
        if mac in self.index_mac:
            hosts.extend(self.index_mac[mac])
            return hosts
        else:
            return None

    def get_client_config_by_duid(self, transaction):
        """
            get host and its information belonging to that DUID
        """
        hosts = list()
        duid = transaction.duid
        if duid in self.index_duid:
            hosts.extend(self.index_duid[duid])
            return hosts
        else:
            return None

    def get_client_config_by_hostname(self, transaction):
        """
            get host and its information by hostname
        """
        hostname = transaction.hostname
        if hostname in self.hosts:
            return [self.hosts[hostname]]
        else:
            return None

    def get_client_config(self, hostname='', client_class='', duid='', address=[], mac=[], host_id=''):
        """
            give back ClientConfig object
        """
        return ClientConfig(hostname=hostname, client_class=client_class, duid=duid, address=address, mac=mac, host_id=host_id)



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import copy

from dns.resolver import (NoAnswer,
                          NoNameservers)

from .config import cfg
from .globals import (dns_query_queue,
                      resolver_query)
from .helpers import decompress_ip6
from .storage import volatile_store


def dns_update(transaction, action='update'):
    """
        update DNS entries on specified nameserver
        at the moment this only works with Bind
        uses all addresses of client if they want to be dynamically updated

        regarding RFC 4704 5. there are 3 kinds of client behaviour for N O S:
        - client wants to update DNS itself -> sends 0 0 0
        - client wants server to update DNS -> sends 0 0 1
        - client wants no server DNS update -> sends 1 0 0
    """
    if transaction.client:
        # if allowed use client supplied hostname, otherwise that from config
        if cfg.DNS_USE_CLIENT_HOSTNAME:
            # hostname from transaction
            hostname = transaction.hostname
        else:
            # hostname from client info built from configuration
            hostname = transaction.client.hostname

        # if address should be updated in DNS update it
        for a in transaction.client.addresses:
            if a.DNS_UPDATE and hostname != '' and a.VALID:
                if cfg.DNS_IGNORE_CLIENT or transaction.dns_s == 1:
                    # put query into DNS query queue
                    dns_query_queue.put((action, hostname, a))
        return True
    else:
        return False


def dns_delete(transaction, address='', action='release'):
    """
        delete DNS entries on specified nameserver
        at the moment this only works with ISC Bind
    """
    hostname, duid, mac, iaid = volatile_store.get_host_lease(address)

    # if address should be updated in DNS update it
    # local flag to check if address should be deleted from DNS
    delete = False

    for a in list(cfg.ADDRESSES.values()):
        # if there is any address type which prototype matches use its DNS ZONE
        if a.matches_prototype(address):
            # kind of RCF-compliant security measure - check if hostname and DUID from transaction fits them of store
            if duid == transaction.duid and \
               iaid == transaction.iaid:
                delete = True
                # also check MAC address if MAC counts in general - not RFCish
                if 'mac' in cfg.IDENTIFICATION:
                    if not mac == transaction.mac:
                        delete = False

            if hostname != '' and delete:
                # use address from address types as template for the real
                # address to be deleted from DNS
                dns_address = copy.copy(a)
                dns_address.ADDRESS = address
                # put query into DNS query queue
                dns_query_queue.put((action, hostname, dns_address))
            # enough
            break


def get_ip_from_dns(hostname):
    """
        Get IPv6 address from DNS for address category 'dns'
    """
    try:
        answer = resolver_query.query(hostname, 'AAAA')
        return decompress_ip6(answer.rrset.to_text().split(' ')[-1])
    except NoAnswer:
        return False
    except NoNameservers:
        return False



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

from binascii import hexlify
from socket import (AF_INET6,
                    inet_pton)

from dhcpy6d import collected_macs
from dhcpy6d.client import Client
from dhcpy6d.config import cfg
from dhcpy6d.constants import CONST
from dhcpy6d.helpers import colonify_ip6
from dhcpy6d.options import OptionTemplate


class Option(OptionTemplate):
    """
    Option 4 + 5 Identity Association for Temporary Address
    """
    def build(self, transaction=None, **kwargs):
        # dummy empty defaults
        response_string_part = ''
        options_answer_part = None

        # check if MAC of LLIP is really known
        if transaction.client_llip in collected_macs or cfg.IGNORE_MAC:
            # collect client information
            if transaction.client is None:
                transaction.client = Client(transaction)

            if 'addresses' in cfg.CLASSES[transaction.client.client_class].ADVERTISE and \
                    CONST.OPTION.IA_TA in transaction.ia_options:
                # check if only a short NoAddrAvail answer or none at all ist t be returned
                if not transaction.answer == 'normal':
                    if transaction.answer == 'noaddress':
                        # Option 13 Status Code Option - statuscode is 2: 'No Addresses available'
                        response_string_part = self.convert_to_string(CONST.OPTION.STATUS_CODE,
                                                                      f'{CONST.STATUS.NO_ADDRESSES_AVAILABLE:04x}')
                        # clean client addresses which not be deployed anyway
                        transaction.client.addresses[:] = []
                        # options in answer to be logged
                        options_answer_part = CONST.OPTION.STATUS_CODE
                else:
                    # if client could not be built because of database problems send
                    # status message back
                    if transaction.client:
                        # embed option 5 into option 4 - several if necessary
                        ia_addresses = ''
                        try:
                            for address in transaction.client.addresses:
                                if address.IA_TYPE == 'ta':
                                    ipv6_address = hexlify(inet_pton(AF_INET6, colonify_ip6(address.ADDRESS))).decode()
                                    # if a transaction consists of too many requests from client -
                                    # - might be caused by going wild Windows clients -
                                    # reset all addresses with lifetime 0
                                    # lets start with maximal transaction count of 10
                                    if transaction.counter < 10:
                                        preferred_lifetime = f'{int(address.PREFERRED_LIFETIME):08x}'
                                        valid_lifetime = f'{int(address.VALID_LIFETIME):08x}'
                                    else:
                                        preferred_lifetime = '00000000'
                                        valid_lifetime = '00000000'
                                    ia_addresses += self.convert_to_string(CONST.OPTION.IAADDR,
                                                                           ipv6_address +
                                                                           preferred_lifetime +
                                                                           valid_lifetime)
                            if ia_addresses != '':
                                response_string_part = self.convert_to_string(CONST.OPTION.IA_TA,
                                                                              transaction.iaid +
                                                                              ia_addresses)
                            # options in answer to be logged
                            options_answer_part = CONST.OPTION.IA_TA
                        except:
                            # Option 13 Status Code Option - statuscode is 2: 'No Addresses available'
                            response_string_part = self.convert_to_string(CONST.OPTION.STATUS_CODE,
                                                                          f'{CONST.STATUS.NO_ADDRESSES_AVAILABLE:04x}')
                            # options in answer to be logged
                            options_answer_part = CONST.OPTION.STATUS_CODE
                    else:
                        # Option 13 Status Code Option - statuscode is 2: 'No Addresses available'
                        response_string_part = self.convert_to_string(CONST.OPTION.STATUS_CODE,
                                                                      f'{CONST.STATUS.NO_ADDRESSES_AVAILABLE:04x}')
                        # options in answer to be logged
                        options_answer_part = CONST.OPTION.STATUS_CODE

        return response_string_part, options_answer_part

    def initialize(self, transaction=None, option=None, **kwargs):
        """
        IA TA addresses of client
        """
        for payload in option:
            transaction.iaid = payload[0:8]
            transaction.iat1 = int(payload[8:16], 16)
            transaction.iat2 = int(payload[16:24], 16)

            # addresses given by client if any
            for a in range(len(payload[32:]) // 44):
                address = payload[32:][(a * 56):(a * 56) + 32]
                # in case an address is asked for twice by one host ignore the twin
                # sometimes address seems to be EMPTY???
                if address and colonify_ip6(address) and address not in transaction.addresses:
                    transaction.addresses.append(address)
        transaction.ia_options.append(CONST.OPTION.IA_TA)



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

from dhcpy6d.options import OptionTemplate


class Option(OptionTemplate):
    """
    Option 1 Client Identifier Option
    """
    def initialize(self, transaction=None, option=None, **kwargs):
        transaction.duid = option
        # See https://github.com/HenriWahl/dhcpy6d/issues/25 and DUID type is not used at all so just remove it
        # self.DUIDType = int(options[1][0:4], 16)
        # # DUID-EN can be retrieved from DUID
        # if self.DUIDType == 2:
        #     # some HP printers seem to produce pretty bad requests, thus some cleaning is necessary
        #     # e.g. '1 1 1 00020000000b0026b1f72a49' instead of '00020000000b0026b1f72a49'
        #     self.DUID_EN = int(options[1].split(' ')[-1][4:12], 16)



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

from binascii import hexlify
from socket import (AF_INET6,
                    inet_pton)

from dhcpy6d.config import cfg
from dhcpy6d.options import OptionTemplate


class Option(OptionTemplate):
    """
    Option 23 DNS recursive name server
    """
    def build(self, transaction=None, **kwargs):
        # dummy empty defaults
        response_string_part = ''
        options_answer_part = None
        # should not be necessary to check if transactions.client exists but there are
        # crazy clients out in the wild which might become silent this way
        if transaction.client:
            if len(cfg.CLASSES[transaction.client.client_class].NAMESERVER) > 0:
                nameserver = b''
                for ns in cfg.CLASSES[transaction.client.client_class].NAMESERVER:
                    nameserver += inet_pton(AF_INET6, ns)
                response_string_part = self.convert_to_string(self.number, hexlify(nameserver).decode())
                options_answer_part = self.number
            elif len(cfg.NAMESERVER) > 0:
                # in case several nameservers are given convert them all and add them
                nameserver = b''
                for ns in cfg.NAMESERVER:
                    nameserver += inet_pton(AF_INET6, ns)
                response_string_part = self.convert_to_string(self.number, hexlify(nameserver).decode())
                options_answer_part = self.number
        return response_string_part, options_answer_part



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import binascii
from copy import deepcopy
import socket
import socketserver
import sys
import traceback

from .client import Client
from .config import cfg
from .constants import CONST
from .domain import (dns_delete,
                     dns_update)
from .globals import (collected_macs,
                      DUMMY_MAC,
                      IA_OPTIONS,
                      requests,
                      requests_blacklist,
                      timer,
                      transactions)
from .helpers import (build_option,
                      colonify_ip6,
                      decompress_ip6,
                      LOCALHOST,
                      LOCALHOST_INTERFACES)
from .log import log
from .macs import collect_macs
from .options import OPTIONS
from .route import modify_route
from .storage import (config_store,
                      volatile_store)
from .transaction import Transaction


class Request:
    """
        to be stored in requests dictionary to log client requests to be able to find brute force clients
    """

    def __init__(self, client):
        self.client = client
        self.count = 1
        self.timestamp = timer.time


class RequestHandler(socketserver.DatagramRequestHandler):
    """
        manage all incoming datagrams, builds clients from config and previous leases
    """
    # empty dummy handler
    response = ''
    # most messages are no locally generated control messages
    is_control_message = False

    def handle(self):
        """
        request handling happens here
        """
        # empty dummy handler
        self.response = ''

        # raw address+interface, used for requests monitoring
        client_address = deepcopy(self.client_address[0].split('%')[0])
        try:
            interface = socket.if_indextoname(self.client_address[3])
        except OSError:
            # the interface index is 0 if sent to localhost -
            # even if 'lo' has the index 1
            interface = ''

        # avoid processing requests of unknown clients which cannot be found in the neighbor cache table
        # only makes sense if classes are not ignored and thus the neighbor cache is used
        if not cfg.IGNORE_MAC and cfg.IGNORE_UNKNOWN_CLIENTS:
            if client_address in requests_blacklist:
                return False

        # check if we are limiting requests
        if cfg.REQUEST_LIMIT:
            if cfg.REQUEST_LIMIT_IDENTIFICATION == 'llip':
                # avoid further processing if client is known to be bad
                if client_address in requests_blacklist:
                    return False
                # add client to requests tracker if not known, otherwise raise counter
                if client_address not in requests:
                    requests[client_address] = Request(client_address)
                else:
                    requests[client_address].count += 1
            # otherwise a MAC address
            else:
                llip = decompress_ip6(client_address)
                if llip in collected_macs:
                    mac = deepcopy(collected_macs[llip].mac)
                    if mac in requests_blacklist:
                        return False
                    # add client to requests tracker if not known, otherwise raise counter
                    if mac not in requests:
                        requests[mac] = Request(mac)
                    else:
                        requests[mac].count += 1
                del llip
        try:
            # convert raw message into ascii-bytes
            raw_bytes = binascii.hexlify(self.request[0]).decode()

            # local connection is a control message
            # for BSD there might be different localhost addresses
            if client_address == LOCALHOST and interface in LOCALHOST_INTERFACES:
                self.is_control_message = True

            # do nothing if interface is not configured
            if interface not in cfg.INTERFACE and not self.is_control_message:
                return False

            # bad or too short message is thrown away
            if not len(raw_bytes) > 8:
                pass
            elif self.is_control_message:
                self.control_message(raw_bytes)
            else:
                message_type = int(raw_bytes[0:2], 16)
                transaction_id = raw_bytes[2:8]
                raw_bytes_options = raw_bytes[8:]
                options = {}
                while len(raw_bytes_options) > 0:
                    # option type and length are 2 bytes each
                    option = int(raw_bytes_options[0:4], 16)
                    length = int(raw_bytes_options[4:8], 16)
                    # *2 because 2 bytes make 1 char
                    value = raw_bytes_options[8:8 + length * 2]

                    # Microsoft behaves a little bit different than the other
                    # clients - in RENEW and REBIND request multiple addresses of an
                    # IAID are not requested all in one option type 3 but
                    # come in several options of type 3 what leads to some confusion
                    # so allow all IA_OPTION (NA, TA, PD) to exist more than once
                    # and create them as list
                    if option not in IA_OPTIONS:
                        options[option] = value
                    else:
                        if option in options:
                            # if options list already exists append value
                            options[option].append(value)
                        else:
                            # otherwise create list and put value in
                            options[option] = [value]

                    # cut off bytes worked on
                    raw_bytes_options = raw_bytes_options[8 + length * 2:]

                # only valid messages will be processed
                if message_type in CONST.MESSAGE_DICT:
                    client_llip = decompress_ip6(client_address)
                    # 2. create Transaction object if not yet done
                    identifier = client_llip + transaction_id
                    if identifier not in transactions:
                        transactions[identifier] = Transaction(transaction_id,
                                                               client_llip,
                                                               interface,
                                                               message_type,
                                                               options)
                        # shortcut to transactions[identifier]
                        transaction = transactions[identifier]
                        # add client MAC address to transaction object
                        if transaction.client_llip in collected_macs and not cfg.IGNORE_MAC:
                            transaction.mac = collected_macs[transaction.client_llip].mac
                    else:
                        # shortcut to transactions[identifier]
                        transaction = transactions[identifier]
                        transaction.timestamp = timer.time
                        transaction.last_message_received_type = message_type

                    # log incoming messages
                    log.info(f'{CONST.MESSAGE_DICT[message_type]} | '
                             f'transaction: {transaction.id}{transaction.get_options_string()}')

                    # 3. answer requests
                    # check if client sent a valid DUID (alphanumeric)
                    if transaction.duid.isalnum():
                        # if request was not addressed to multicast do nothing but logging
                        if transaction.interface == '':
                            log.info(f'transaction: {transaction.id} | Multicast necessary '
                                     f'but message came from {colonify_ip6(transaction.client_llip)}')
                            # reset transaction counter
                            transaction.counter = 0
                        else:
                            # client will get answer if its LLIP & MAC is known
                            if transaction.client_llip not in collected_macs:
                                if not cfg.IGNORE_MAC:
                                    # complete MAC collection - will make most sense on Linux
                                    # and its native neighborcache access
                                    collect_macs(timer.time)

                                    # when still no trace of the client in neighbor cache then send silly signal back
                                    if transaction.client_llip not in collected_macs:
                                        # if not known send status code option failure to get
                                        # LLIP/MAC mapping from neighbor cache
                                        # status code 'Success' sounds silly but works best
                                        self.build_response(CONST.MESSAGE.REPLY,
                                                            transaction,
                                                            [CONST.OPTION.STATUS_CODE],
                                                            status=CONST.STATUS.SUCCESS)
                                        # complete MAC collection
                                        collect_macs(timer.time)
                                        # if client cannot be found in collected MACs
                                        if transaction.client_llip not in collected_macs:
                                            if cfg.IGNORE_UNKNOWN_CLIENTS and client_address in requests:
                                                if requests[client_address].count > 1:
                                                    requests_blacklist[client_address] = Request(client_address)
                                                    log.info(f"Blacklisting unknown client {client_address}")
                                                    return False

                                    # try to add client MAC address to transaction object
                                    try:
                                        transaction.mac = collected_macs[transaction.client_llip].mac
                                    except KeyError:
                                        # MAC not yet found :-(
                                        if cfg.LOG_MAC_LLIP:
                                            log.info(f'transaction: {transaction.id} | mac address for '
                                                     f'llip {colonify_ip6(transaction.client_llip)} unknown')

                            # if finally there is some info about the client or MACs
                            # it plays no role try to answer the request
                            if transaction.client_llip in collected_macs or cfg.IGNORE_MAC:
                                if not cfg.IGNORE_MAC:
                                    if transaction.mac == DUMMY_MAC:
                                        transaction.mac = collected_macs[transaction.client_llip].mac

                                # ADVERTISE
                                # if last request was a SOLICIT send an ADVERTISE (type 2) back
                                if transaction.last_message_received_type == CONST.MESSAGE.SOLICIT \
                                        and not transaction.rapid_commit:
                                    # preference option (7) is for free
                                    self.build_response(CONST.MESSAGE.ADVERTISE,
                                                        transaction,
                                                        transaction.ia_options +
                                                        [CONST.OPTION.PREFERENCE] +
                                                        transaction.options_request)

                                    # store leases for addresses and lock advertised address
                                    volatile_store.store(deepcopy(transaction), timer.time)

                                # REQUEST
                                # if last request was a REQUEST (type 3) send a REPLY (type 7) back
                                elif transaction.last_message_received_type == CONST.MESSAGE.REQUEST or \
                                        (transaction.last_message_received_type == CONST.MESSAGE.SOLICIT and
                                         transaction.rapid_commit):
                                    # preference option (7) is for free
                                    # if RapidCommit was set give it back
                                    if not transaction.rapid_commit:
                                        self.build_response(CONST.MESSAGE.REPLY,
                                                            transaction,
                                                            transaction.ia_options +
                                                            [CONST.OPTION.PREFERENCE] +
                                                            transaction.options_request)
                                    else:
                                        self.build_response(CONST.MESSAGE.REPLY,
                                                            transaction,
                                                            transaction.ia_options +
                                                            [CONST.OPTION.PREFERENCE] +
                                                            [CONST.OPTION.RAPID_COMMIT] + transaction.options_request)
                                    # store leases for addresses
                                    volatile_store.store(deepcopy(transaction), timer.time)

                                    # run external script for setting a route to the delegated prefix
                                    if CONST.OPTION.IA_PD in transaction.ia_options:
                                        modify_route(transaction, 'up')

                                    if cfg.DNS_UPDATE:
                                        dns_update(transaction)

                                # CONFIRM
                                # if last request was a CONFIRM (4) send a REPLY (type 7) back
                                # Due to problems with different clients they will get a not-available-reply
                                # but the next ADVERTISE will offer them the last known and still active
                                # lease. This makes sense in case of fixed MAC-based, addresses, ranges and
                                # ID-based addresses, Random addresses will be recalculated
                                elif transaction.last_message_received_type == CONST.MESSAGE.CONFIRM:
                                    # the RFC 3315 is a little bit confusing regarding CONFIRM
                                    # messages so it won't hurt to simply let the client
                                    # solicit addresses again via answering 'NotOnLink'
                                    # thus client is forced in every case to solicit a new address which
                                    # might as well be the old one or a new if prefix has changed
                                    self.build_response(CONST.MESSAGE.REPLY,
                                                        transaction,
                                                        [CONST.OPTION.STATUS_CODE],
                                                        status=CONST.STATUS.PREFIX_NOT_APPROPRIATE_FOR_LINK)

                                # RENEW
                                # if last request was a RENEW (type 5) send a REPLY (type 7) back
                                elif transaction.last_message_received_type == CONST.MESSAGE.RENEW:
                                    self.build_response(CONST.MESSAGE.REPLY,
                                                        transaction,
                                                        transaction.ia_options +
                                                        [CONST.OPTION.PREFERENCE] +
                                                        transaction.options_request)
                                    # store leases for addresses
                                    volatile_store.store(deepcopy(transaction), timer.time)
                                    if cfg.DNS_UPDATE:
                                        dns_update(transaction)

                                # REBIND
                                # if last request was a REBIND (type 6) send a REPLY (type 7) back
                                elif transaction.last_message_received_type == CONST.MESSAGE.REBIND:
                                    self.build_response(CONST.MESSAGE.REPLY,
                                                        transaction,
                                                        transaction.ia_options +
                                                        [CONST.OPTION.PREFERENCE] +
                                                        transaction.options_request)
                                    # store leases for addresses
                                    volatile_store.store(deepcopy(transaction), timer.time)

                                # RELEASE
                                # if last request was a RELEASE (type 8) send a REPLY (type 7) back
                                elif transaction.last_message_received_type == CONST.MESSAGE.RELEASE:
                                    #  build client to be able to delete it from DNS
                                    if transaction.client is None:
                                        transaction.client = Client(transaction)
                                    if cfg.DNS_UPDATE:
                                        for a in transaction.addresses:
                                            dns_delete(transaction, address=a, action='release')
                                    for a in transaction.addresses:
                                        # free lease
                                        volatile_store.release_lease(a, timer.time)
                                    for p in transaction.prefixes:
                                        # free prefix - without length
                                        volatile_store.release_prefix(p.split('/')[0], timer.time)
                                        # delete route to formerly requesting client
                                        modify_route(transaction, 'down')
                                    # send status code option (type 13) with success (type 0)
                                    self.build_response(CONST.MESSAGE.REPLY,
                                                        transaction,
                                                        [CONST.OPTION.STATUS_CODE],
                                                        status=CONST.STATUS.SUCCESS)

                                # DECLINE
                                # if last request was a DECLINE (type 9) send a REPLY (type 7) back
                                elif transaction.last_message_received_type == CONST.MESSAGE.DECLINE:
                                    # maybe has to be refined - now only a status code 'NoBinding' is answered
                                    self.build_response(CONST.MESSAGE.REPLY,
                                                        transaction,
                                                        [CONST.OPTION.STATUS_CODE],
                                                        status=CONST.STATUS.NO_BINDING)

                                # INFORMATION-REQUEST
                                # if last request was an INFORMATION-REQUEST (type 11) send a REPLY (type 7) back
                                elif transaction.last_message_received_type == CONST.MESSAGE.INFORMATION_REQUEST:
                                    self.build_response(CONST.MESSAGE.REPLY,
                                                        transaction,
                                                        transaction.options_request)

                                # general error - statuscode 1 'Failure'
                                else:
                                    # send Status Code Option (type 13) with status code 'UnspecFail'
                                    self.build_response(CONST.MESSAGE.REPLY,
                                                        transaction,
                                                        [CONST.OPTION.STATUS_CODE],
                                                        status=CONST.STATUS.FAILURE)

                    # count requests of transaction
                    # if there will be too much something went wrong
                    # may be evaluated to reset the whole transaction
                    transaction.counter += 1

        except Exception as err:
            traceback.print_exc(file=sys.stdout)
            sys.stdout.flush()
            log.error(f'handle(): {str(err)} | caused by: {client_address} | transaction: {transaction.id}')
            return None

    def build_response(self, message_type_response, transaction, options_request, status=0):
        """
            creates answer and puts it into self.handler
            arguments:
                message_type_response - mostly 2 or 7
                transaction
                option_request
                status - mostly 0 (OK)
            handler will be sent by self.finish()
        """
        try:
            # should be asked before any responses are built
            if transaction.answer == 'none':
                self.response = ''
                return None

            # Header
            # handler type + transaction id
            response_string = f'{message_type_response:02x}'
            response_string += transaction.id

            # these options are always useful
            # Option 1 client identifier
            response_string += build_option(CONST.OPTION.CLIENTID, transaction.duid)
            # Option 2 server identifier
            response_string += build_option(CONST.OPTION.SERVERID, cfg.SERVERDUID)

            # list of options in answer to be logged
            options_answer = []

            # build all requested options if they are handled
            for number in options_request:
                if number in OPTIONS:
                    try:
                        response_string_part, options_answer_part = OPTIONS[number].build(transaction=transaction,
                                                                                          status=status)
                        response_string += response_string_part
                        if options_answer_part:
                            options_answer.append(options_answer_part)
                    except Exception:
                        traceback.print_exc(file=sys.stdout)
                        sys.stdout.flush()

            # if databases are not connected send error to client
            # if not (config_store.connected == volatile_store.connected == True):
            if not config_store.connected and not volatile_store.connected:
                # mark database errors - every database may add its error
                db_error = []
                if not config_store.connected:
                    db_error.append('config')
                    config_store.db_connect()
                if not volatile_store.connected:
                    db_error.append('volatile')
                    volatile_store.db_connect()

                # create error handler - headers have to be recreated because
                # problems may have arisen while processing and these information
                # is not valid anymore
                # handler type + transaction id
                response_string = f'{CONST.MESSAGE.REPLY:02x}'
                response_string += transaction.id

                # always of interest
                # option 1 client identifier
                response_string += build_option(CONST.OPTION.CLIENTID,
                                                transaction.duid)
                # option 2 server identifier
                response_string += build_option(CONST.OPTION.SERVERID,
                                                cfg.SERVERDUID)

                # Option 13 Status Code Option - statuscode is 2: 'No Addresses available'
                response_string += build_option(CONST.OPTION.STATUS_CODE,
                                                f'{CONST.STATUS.NO_ADDRESSES_AVAILABLE:04x}')

                log.error(f'{CONST.MESSAGE_DICT[message_type_response]} | '
                          f'transaction: {transaction.id} | '
                          f'DatabaseError: {" ".join(db_error)}')
            else:
                # log handler
                # if transaction.client is not None:
                if transaction.client:
                    if len(transaction.client.addresses) == 0 and \
                            len(transaction.client.prefixes) == 0 and \
                            transaction.answer == 'normal' and \
                            transaction.last_message_received_type in [CONST.MESSAGE.SOLICIT,
                                                                       CONST.MESSAGE.REQUEST,
                                                                       CONST.MESSAGE.RENEW,
                                                                       CONST.MESSAGE.REBIND]:
                        # create error handler - headers have to be recreated because
                        # problems may have arisen while processing and these information
                        # is not valid anymore
                        # handler type + transaction id
                        response_string = f'{CONST.MESSAGE.REPLY:02x}'
                        response_string += transaction.id

                        # always of interest
                        # option 1 client identifier
                        response_string += build_option(CONST.OPTION.CLIENTID,
                                                        transaction.duid)
                        # option 2 server identifier
                        response_string += build_option(CONST.OPTION.SERVERID,
                                                        cfg.SERVERDUID)

                        # Option 13 Status Code Option - statuscode is 2: 'No Addresses available'
                        response_string += build_option(CONST.OPTION.STATUS_CODE,
                                                        f'{CONST.STATUS.NO_ADDRESSES_AVAILABLE:04x}')
                        # options in answer to be logged
                        options_answer.append(CONST.OPTION.STATUS_CODE)

                        # log warning message about unavailable addresses
                        log.warning(f'REPLY | no addresses or prefixes available | '
                                    f'transaction: {transaction.id} | '
                                    f'client_llip: {colonify_ip6(transaction.client_llip)}')

                    elif CONST.OPTION.IA_NA in options_request or \
                            CONST.OPTION.IA_TA in options_request or \
                            CONST.OPTION.IA_PD in options_request or \
                            CONST.OPTION.STATUS_CODE in options_request:
                        options_answer = sorted(options_answer)
                        log.info(f'{CONST.MESSAGE_DICT[message_type_response]} | '
                                 f'transaction: {transaction.id} | '
                                 f'options: {options_answer} {transaction.client.get_options_string()}')
                    else:
                        print(options_request)
                        log.info('what else should I do?')
                else:
                    options_answer = sorted(options_answer)
                    log.info(f'{CONST.MESSAGE_DICT[message_type_response]} | '
                             f'transaction: {transaction.id} | '
                             f'options: {options_answer}')
            # handler
            self.response = binascii.unhexlify(response_string)

        except Exception as err:
            traceback.print_exc(file=sys.stdout)
            sys.stdout.flush()
            log.error('handler: ' + str(err))
            # clear any handler
            self.response = ''

            return None

    def finish(self):
        """
        send handler from self.handler
        """
        # send only if there is anything to send
        if cfg.REALLY_DO_IT:
            if len(self.response) > 0:
                self.socket.sendto(self.response, self.client_address)
        else:
            log.error("Nothing sent - please set 'really_do_it = yes' in config file or as command line option.")

    @staticmethod
    def control_message(raw_bytes):
        """
        execute commands sent in by control message
        @staticmethod proposed by PyCharm
        """
        control_message = binascii.unhexlify(raw_bytes)
        control_message_fragments = control_message.decode().split(' ')
        # clean message
        control_message_clean = list()
        for count in range(len(control_message_fragments)):
            if control_message_fragments[count] != '':
                control_message_clean.append(control_message_fragments[count])
        command = control_message_clean[0]
        arguments = control_message_clean[1:]

        # change dynamic prefix
        if command == 'prefix' and len(arguments) == 1:
            cfg.PREFIX = arguments[0]
            volatile_store.store_dynamic_prefix(cfg.PREFIX)
            # apply dynamic prefix to addresses and prefixes
            for a in cfg.ADDRESSES:
                cfg.ADDRESSES[a].inject_dynamic_prefix_into_prototype(cfg.PREFIX)
            for p in cfg.PREFIXES:
                cfg.PREFIXES[p].inject_dynamic_prefix_into_prototype(cfg.PREFIX)
        log.info(f'Control message \'{" ".join(control_message_clean)}\' received')



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

from ..config import (Address,
                      cfg,
                      Prefix)
from ..constants import CONST
from ..globals import timer
from ..helpers import (decompress_ip6,
                       decompress_prefix,
                       split_prefix)
from ..storage import volatile_store

from .parse_pattern import parse_pattern_address


def reuse_lease(client=None, client_config=None, transaction=None):
    """
    Reuse already existing lease information
    """
    if client_config is not None:
        # give client hostname
        client.hostname = client_config.HOSTNAME
        client.client_class = client_config.CLASS
        # apply answer type of client to transaction - useful if no answer or no address available is configured
        transaction.answer = cfg.CLASSES[client.client_class].ANSWER
    else:
        # use default class if host is unknown
        client.hostname = transaction.hostname
        client.client_class = 'default_' + transaction.interface
        # apply answer type of client to transaction - useful if no answer or no address available is configured
        transaction.answer = cfg.CLASSES[client.client_class].ANSWER

    if 'addresses' in cfg.CLASSES[client.client_class].ADVERTISE and \
            (CONST.OPTION.IA_NA or CONST.OPTION.IA_TA) in transaction.ia_options:
        for address in transaction.addresses:
            # check_lease returns hostname, address, type, category, ia_type, class, preferred_until of leased address
            answer = volatile_store.check_lease(address, transaction)
            if answer:
                if len(answer) > 0:
                    for item in answer:
                        a = dict(list(
                            zip(('hostname', 'address', 'type', 'category', 'ia_type', 'class', 'preferred_until'),
                                item)))
                        # if lease exists but no configured client set class to default
                        if client_config is None:
                            client.hostname = transaction.hostname
                            client.client_class = 'default_' + transaction.interface
                        # check if address type of lease still exists in configuration
                        # and if request interface matches that of class
                        if a['class'] in cfg.CLASSES and client.client_class == a['class'] and \
                                transaction.interface in cfg.CLASSES[client.client_class].INTERFACE:
                            # type of address must be defined in addresses for this class
                            # or fixed/dns - in which case it is not class related
                            if a['type'] in cfg.CLASSES[a['class']].ADDRESSES or a['type'] in ['fixed']:
                                # flag for lease usage
                                use_lease = True
                                # test lease validity against address prototype pattern only if not fixed or from DNS
                                if not a['category'] in ['fixed', 'dns']:
                                    # test if address matches pattern
                                    for identification in range(len(address)):
                                        if address[identification] != cfg.ADDRESSES[a['type']].PROTOTYPE[
                                            identification] and \
                                                cfg.ADDRESSES[a['type']].PROTOTYPE[identification] != 'x':
                                            use_lease = False
                                            break
                                elif a['category'] == 'fixed' and client_config.ADDRESS is not None:
                                    if address not in client_config.ADDRESS:
                                        use_lease = False
                                elif a['category'] == 'dns':
                                    use_lease = False

                                # only use lease if it still matches prototype
                                if use_lease:
                                    # when category is range, test if it still applies
                                    if a['category'] == 'range':
                                        # borrowed from parse_pattern_address to find out if lease is still in
                                        # a meanwhile maybe changed range
                                        range_from, range_to = cfg.ADDRESSES[a['type']].RANGE.split('-')

                                        # correct possible misconfiguration
                                        if len(range_from) < 4:
                                            range_from = '0' * (4 - len(range_from)) + range_from
                                        if len(range_to) < 4:
                                            range_to = '0' * (4 - len(range_to)) + range_to
                                        if range_from > range_to:
                                            range_from, range_to = range_to, range_from
                                        # if lease is still inside range boundaries use it
                                        if range_from <= address[28:].lower() < range_to:
                                            # build IA partly of leases db, partly of config db
                                            ia = Address(address=a['address'],
                                                         atype=a['type'],
                                                         preferred_lifetime=cfg.ADDRESSES[a['type']].PREFERRED_LIFETIME,
                                                         valid_lifetime=cfg.ADDRESSES[a['type']].VALID_LIFETIME,
                                                         category=a['category'],
                                                         ia_type=a['ia_type'],
                                                         aclass=a['class'],
                                                         dns_update=cfg.ADDRESSES[a['type']].DNS_UPDATE,
                                                         dns_zone=cfg.ADDRESSES[a['type']].DNS_ZONE,
                                                         dns_rev_zone=cfg.ADDRESSES[a['type']].DNS_REV_ZONE,
                                                         dns_ttl=cfg.ADDRESSES[a['type']].DNS_TTL)
                                            client.addresses.append(ia)

                                    # de-preferred random address has to be deleted and replaced
                                    elif a['category'] == 'random' and timer.time > a['preferred_until']:
                                        # create new random address if old one is depreferred
                                        random_address = parse_pattern_address(cfg.ADDRESSES[a['type']],
                                                                               client_config,
                                                                               transaction)
                                        # create new random address if old one is de-preferred
                                        # do not wait until it is invalid
                                        if random_address is not None:
                                            ia = Address(address=random_address,
                                                         ia_type=cfg.ADDRESSES[a['type']].IA_TYPE,
                                                         preferred_lifetime=cfg.ADDRESSES[a['type']].PREFERRED_LIFETIME,
                                                         valid_lifetime=cfg.ADDRESSES[a['type']].VALID_LIFETIME,
                                                         category='random',
                                                         aclass=cfg.ADDRESSES[a['type']].CLASS,
                                                         atype=cfg.ADDRESSES[a['type']].TYPE,
                                                         dns_update=cfg.ADDRESSES[a['type']].DNS_UPDATE,
                                                         dns_zone=cfg.ADDRESSES[a['type']].DNS_ZONE,
                                                         dns_rev_zone=cfg.ADDRESSES[a['type']].DNS_REV_ZONE,
                                                         dns_ttl=cfg.ADDRESSES[a['type']].DNS_TTL)
                                            client.addresses.append(ia)
                                            # set de-preferred address invalid
                                            client.addresses.append(Address(address=a['address'],
                                                                            valid=False,
                                                                            preferred_lifetime=0,
                                                                            valid_lifetime=0))

                                    else:
                                        # build IA partly of leases db, partly of config db
                                        ia = Address(address=a['address'],
                                                     atype=a['type'],
                                                     preferred_lifetime=cfg.ADDRESSES[a['type']].PREFERRED_LIFETIME,
                                                     valid_lifetime=cfg.ADDRESSES[a['type']].VALID_LIFETIME,
                                                     category=a['category'],
                                                     ia_type=a['ia_type'],
                                                     aclass=a['class'],
                                                     dns_update=cfg.ADDRESSES[a['type']].DNS_UPDATE,
                                                     dns_zone=cfg.ADDRESSES[a['type']].DNS_ZONE,
                                                     dns_rev_zone=cfg.ADDRESSES[a['type']].DNS_REV_ZONE,
                                                     dns_ttl=cfg.ADDRESSES[a['type']].DNS_TTL)
                                        client.addresses.append(ia)

        # important indent here, has to match for...addresses-loop!
        # look for addresses in transaction that are invalid and add them
        # to client addresses with flag invalid and a RFC-compliant lifetime of 0
        for a in set(transaction.addresses).difference(
                [decompress_ip6(x.ADDRESS) for x in client.addresses]):
            client.addresses.append(Address(address=a,
                                            valid=False,
                                            preferred_lifetime=0,
                                            valid_lifetime=0))

    if 'prefixes' in cfg.CLASSES[client.client_class].ADVERTISE and \
            CONST.OPTION.IA_PD in transaction.ia_options:
        for prefix in transaction.prefixes:
            # split prefix of prefix from length, separated by /
            prefix_prefix, prefix_length = split_prefix(prefix)

            # check_prefix returns hostname, prefix, length, type, category, class, preferred_until of leased address
            answer = volatile_store.check_prefix(prefix_prefix, prefix_length, transaction)

            if answer:
                if len(answer) > 0:
                    for item in answer:
                        p = dict(list(
                            zip(('hostname', 'prefix', 'length', 'type', 'category', 'class', 'preferred_until'),
                                item)))
                        # if lease exists but no configured client set class to default
                        if client_config is None:
                            client.hostname = transaction.hostname
                            client.client_class = 'default_' + transaction.interface
                        # check if address type of lease still exists in configuration
                        # and if request interface matches that of class
                        if p['class'] in cfg.CLASSES and client.client_class == p['class'] and \
                                transaction.interface in cfg.CLASSES[client.client_class].INTERFACE:
                            # type of address must be defined in addresses for this class
                            # or fixed/dns - in which case it is not class related
                            if p['type'] in cfg.CLASSES[p['class']].PREFIXES:
                                # flag for lease usage
                                use_lease = True
                                # test if prefix matches pattern
                                for identification in range(len(prefix_prefix)):
                                    if prefix_prefix[identification] != cfg.PREFIXES[p['type']].PROTOTYPE[
                                        identification] and \
                                            cfg.PREFIXES[p['type']].PROTOTYPE[identification] != 'x':
                                        use_lease = False
                                        break
                                # only use prefix if it still matches prototype
                                if use_lease:
                                    # when category is range, test if it still applies
                                    if p['category'] == 'range':
                                        # borrowed from parse_pattern_prefix to find out if lease
                                        # is still in a meanwhile maybe changed range
                                        range_from, range_to = cfg.PREFIXES[p['type']].RANGE.split('-')

                                        # correct possible misconfiguration
                                        if len(range_from) < 4:
                                            range_from = '0' * (4 - len(range_from)) + range_from
                                        if len(range_to) < 4:
                                            range_to = '0' * (4 - len(range_to)) + range_to
                                        if range_from > range_to:
                                            range_from, range_to = range_to, range_from

                                        # contrary to addresses the prefix $range$ part of the pattern is expected
                                        # somewhere at the left part of the pattern
                                        # here the 128 Bit sum up to 32 characters in address/prefix string so
                                        # prefix_range_index has to be calculated as first character of range part of
                                        # prefix - assuming steps of width 4
                                        prefix_range_index = int(cfg.PREFIXES[p['type']].LENGTH) // 4 - 4
                                        # prefix itself has a prefix - the first part of the prefix pattern
                                        prefix_prefix = decompress_ip6(p['prefix'].replace('$range$', '0000'))[
                                                        :prefix_range_index + 4]

                                        # if lease is still inside range boundaries use it
                                        if range_from <= prefix_prefix[
                                                         prefix_range_index:prefix_range_index + 4].lower() < range_to:
                                            # build IA partly of leases db, partly of config db
                                            ia = Prefix(prefix=p['prefix'],
                                                        length=p['length'],
                                                        ptype=p['type'],
                                                        preferred_lifetime=cfg.PREFIXES[p['type']].PREFERRED_LIFETIME,
                                                        valid_lifetime=cfg.PREFIXES[p['type']].VALID_LIFETIME,
                                                        category=p['category'],
                                                        pclass=p['class'],
                                                        route_link_local=cfg.PREFIXES[p['type']].ROUTE_LINK_LOCAL)
                                            client.prefixes.append(ia)
                            # add prefixes which are bound to client to advertised prefixes
                            else:
                                ia = Prefix(prefix=p['prefix'],
                                            length=p['length'],
                                            ptype=p['type'],
                                            preferred_lifetime=cfg.PREFERRED_LIFETIME,
                                            valid_lifetime=cfg.VALID_LIFETIME,
                                            category=p['category'],
                                            pclass=p['class'],
                                            route_link_local=False)
                                client.prefixes.append(ia)

        # important indent here, has to match for...prefixes-loop!
        # look for prefixes in transaction that are invalid and add them
        # to client prefixes with flag invalid and a RFC-compliant lifetime of 0
        if len(client.prefixes) > 0:
            for p in set(transaction.prefixes).difference(
                    [decompress_prefix(x.PREFIX, x.LENGTH) for x in client.prefixes]):
                prefix, length = split_prefix(p)
                client.prefixes.append(Prefix(prefix=prefix,
                                              length=length,
                                              valid=False,
                                              preferred_lifetime=0,
                                              valid_lifetime=0))
                del (prefix, length)

    # given client has been modified successfully
    return True



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import socket
import struct

# DHCPv6
MESSAGE = {1: 'SOLICIT',
           2: 'ADVERTISE',
           3: 'REQUEST',
           4: 'CONFIRM',
           5: 'RENEW',
           6: 'REBIND',
           7: 'REPLY',
           8: 'RELEASE',
           9: 'DECLINE',
           10: 'RECONFIGURE',
           11: 'INFORMATION-REQUEST',
           12: 'RELAY-FORW',
           13: 'RELAY-REPL'}

# see http://www.iana.org/assignments/dhcpv6-parameters/
OPTION = {1: 'CLIENTID',
          2: 'SERVERID',
          3: 'IA_NA',
          4: 'IA_TA',
          5: 'IAADDR',
          6: 'ORO',
          7: 'PREFERENCE',
          8: 'ELAPSED_TIME',
          9: 'RELAY_MSG',
          11: 'AUTH',
          12: 'UNICAST',
          13: 'STATUS_CODE',
          14: 'RAPID_COMMIT',
          15: 'USER_CLASS',
          16: 'VENDOR_CLASS',
          17: 'VENDOR_OPTS',
          18: 'INTERFACE_ID',
          19: 'RECONF_MSG',
          20: 'RECONF_ACCEPT',
          21: 'SIP_SERVER_D',
          22: 'SIP_SERVER_A',
          23: 'DNS_SERVERS',
          24: 'DOMAIN_LIST',
          25: 'IA_PD',
          26: 'IAPREFIX',
          27: 'NIS_SERVERS',
          28: 'NISP_SERVERS',
          29: 'NIS_DOMAIN_NAME',
          30: 'NISP_DOMAIN_NAME',
          31: 'SNTP_SERVERS',
          32: 'INFORMATION_REFRESH_TIME',
          33: 'BCMCS_SERVER_D',
          34: 'BCMCS_SERVER_A',
          36: 'GEOCONF_CIVIC',
          37: 'REMOTE_ID',
          38: 'SUBSCRIBER_ID',
          39: 'CLIENT_FQDN',
          40: 'PANA_AGENT',
          41: 'NEW_POSIX_TIMEZONE',
          42: 'NEW_TZDB_TIMEZONE',
          43: 'ERO',
          44: 'LQ_QUERY',
          45: 'CLIENT_DATA',
          46: 'CLT_TIME',
          47: 'LQ_RELAY_DATA',
          48: 'LQ_CLIENT_LINK',
          49: 'MIP6_HNINF',
          50: 'MIP6_RELAY',
          51: 'V6_LOST',
          52: 'CAPWAP_AC_V6',
          53: 'RELAY_ID',
          54: 'IPv6_Address_MoS',
          55: 'Pv6_FQDN_MoS',
          56: 'NTP_SERVER',
          57: 'V6_ACCESS_DOMAIN',
          58: 'SIP_UA_CS_LIST',
          59: 'BOOTFILE_URL',
          60: 'OPT_BOOTFILE_PARAM',
          61: 'OPTION_CLIENT_ARCH_TYPE'
          }

STATUS = {0: 'Success',
          1: 'Failure',
          2: 'No Addresses available',
          3: 'No Binding',
          4: 'Prefix not appropriate for link',
          5: 'Use Multicast',
          6: 'No Prefix available'}

# see https://tools.ietf.org/html/rfc4578#section-2.1
ARCHITECTURE_TYPE = {0: 'Intel x86PC',
                     1: 'NEC / PC98',
                     2: 'EFI Itanium',
                     3: 'DEC Alpha',
                     4: 'Arc x86',
                     5: 'Intel Lean Client',
                     6: 'EFI IA32',
                     7: 'EFI BC',
                     8: 'EFI Xscale',
                     9: 'EFI x86 - 64'}

# used for NETLINK in get_neighbor_cache_linux() access by Github/vokac
RTM_NEWNEIGH = 28
RTM_DELNEIGH = 29
RTM_GETNEIGH = 30
NLM_F_REQUEST = 1
# Modifiers to GET request
NLM_F_ROOT = 0x100
NLM_F_MATCH = 0x200
NLM_F_DUMP = (NLM_F_ROOT | NLM_F_MATCH)
# NETLINK message is always the same except header seq
MSG = struct.pack('B', socket.AF_INET6)
# always the same length...
MSG_HEADER_LENGTH = 17
# ...type...
MSG_HEADER_TYPE = RTM_GETNEIGH
# ...flags.
MSG_HEADER_FLAGS = (NLM_F_REQUEST | NLM_F_DUMP)
NLMSG_NOOP = 0x1  # /* Nothing.             */
NLMSG_ERROR = 0x2  # /* Error                */
NLMSG_DONE = 0x3  # /* End of a dump        */
NLMSG_OVERRUN = 0x4  # /* Data lost            */

NUD_INCOMPLETE = 0x01
# state of peer
NUD_REACHABLE = 0x02
NUD_STALE = 0x04
NUD_DELAY = 0x08
NUD_PROBE = 0x10
NUD_FAILED = 0x20
NUD_NOARP = 0x40
NUD_PERMANENT = 0x80
NUD_NONE = 0x00

NDA = {
    0: 'NDA_UNSPEC',
    1: 'NDA_DST',
    2: 'NDA_LLADDR',
    3: 'NDA_CACHEINFO',
    4: 'NDA_PROBES',
    5: 'NDA_VLAN',
    6: 'NDA_PORT',
    7: 'NDA_VNI',
    8: 'NDA_IFINDEX',
}
NLMSG_ALIGNTO = 4
NLA_ALIGNTO = 4


# collect most constants in a class for easier handling by calling numeric values via class properties
# at the same time still available with integer keys and string values
class Constants:
    """
    contains various categories of constants
    """
    class Category:
        """
        category containing constants
        'reverting' the dictionary because in certain parts for example the number of an option is referred to by
        its name as property
        """
        def __init__(self, category):
            for key, value in category.items():
                self.__dict__[value.replace('-', '_').replace(' ', '_').replace('/', 'or').upper()] = key

        def keys(self):
            # return key
            return self.__dict__.keys()

    def __init__(self):
        self.MESSAGE = self.Category(MESSAGE)
        self.STATUS = self.Category(STATUS)
        self.OPTION = self.Category(OPTION)
        # needed for logging - use original dict
        self.MESSAGE_DICT = MESSAGE
        # architecture types as dict
        self.ARCHITECTURE_TYPE_DICT = ARCHITECTURE_TYPE


# Add constants for global access
CONST = Constants()



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

from binascii import hexlify
from socket import (AF_INET6,
                    inet_pton)

from dhcpy6d.config import cfg
from dhcpy6d.helpers import convert_dns_to_binary
from dhcpy6d.options import OptionTemplate


class Option(OptionTemplate):
    """
    Option 56 NTP server
    https://tools.ietf.org/html/rfc5908
    """
    def build(self, **kwargs):
        # dummy empty defaults
        response_string_part = ''
        options_answer_part = None
        ntp_server_options = ''

        if len(cfg.NTP_SERVER) > 0:
            for ntp_server_type in list(cfg.NTP_SERVER_DICT.keys()):
                # ntp_server_suboption
                for ntp_server in cfg.NTP_SERVER_DICT[ntp_server_type]:
                    ntp_server_suboption = ''
                    if ntp_server_type == 'SRV':
                        ntp_server_suboption = self.convert_to_string(1, hexlify(inet_pton(AF_INET6, ntp_server)).decode())
                    elif ntp_server_type == 'MC':
                        ntp_server_suboption = self.convert_to_string(2, hexlify(inet_pton(AF_INET6, ntp_server)).decode())
                    elif ntp_server_type == 'FQDN':
                        ntp_server_suboption = self.convert_to_string(3, convert_dns_to_binary(ntp_server))
                    ntp_server_options += ntp_server_suboption
            response_string_part = self.convert_to_string(self.number, ntp_server_options)
            # options in answer to be logged
            options_answer_part = self.number

        return response_string_part, options_answer_part



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

from .config import cfg
from .constants import CONST
from .globals import (DUMMY_IAID,
                      DUMMY_MAC,
                      EMPTY_OPTIONS,
                      IGNORED_LOG_OPTIONS,
                      timer)
from .helpers import (colonify_ip6,
                      combine_prefix_length,
                      split_prefix)
from .options import OPTIONS


class Transaction:
    """
    all data of one transaction, to be collected in Transactions
    """
    def __init__(self, transaction_id, client_llip, interface, message_type, options):
        # Transaction ID
        self.id = transaction_id
        # Link Local IP of client
        self.client_llip = client_llip
        # Interface the request came in
        self.interface = interface
        # MAC address
        self.mac = DUMMY_MAC
        # last message for following the protocol
        self.last_message_received_type = message_type
        # dictionary for options
        self.options_raw = options
        # default dummy OptionsRequest
        self.options_request = list()
        # timestamp to manage/clean transactions
        self.timestamp = timer.time
        # dummy hostname
        self.fqdn = ''
        self.hostname = ''
        # DNS Options for option 39
        self.dns_n = 0
        self.dns_o = 0
        self.dns_s = 0
        # dummy IAID
        self.iaid = DUMMY_IAID
        # dummy IAT1
        self.iat1 = cfg.T1
        # dummy IAT2
        self.iat2 = cfg.T2
        # IA option - NA, TA or PD -> DHCPv6 option 3, 4 or 25
        # to be used in option_requests in Handler.build_response()
        self.ia_options = []
        # Addresses given by client, for example for RENEW or RELEASE requests
        self.addresses = []
        # same with prefixes
        self.prefixes = []
        # might be used against clients that are running wild
        # initial 1 as being increased after handling
        self.counter = 1
        # temporary storage for client configuration from DB config
        # - only used if config comes from DB
        self.client_config_dicts = None
        # client config from config store
        self.client = None
        # Vendor Class Option
        self.vendor_class_en = None
        self.vendor_class_data = ''
        # Rapid Commit flag
        self.rapid_commit = False
        # answer type - take from class definition, one of 'normal', 'noaddress', 'noprefix' or 'none'
        # defaults to 'normal' as this is the main purpose of dhcpy6d
        self.answer = 'normal'
        # default DUID value
        self.duid = ''
        # Elapsed Time - option 8, at least sent by WIDE dhcp6c when requesting delegated prefix
        self.elapsed_time = 0
        # Client architecture type (RFC 5970)
        self.client_architecture = ''
        # Known client architecture type (RFC 4578) (e.g. EFI x86 - 64)
        self.known_client_architecture = ''
        # UserClass (https://tools.ietf.org/html/rfc3315#section-22.15)
        self.user_class = ''

        # if the options have some treatment for transactions just apply it if there is an defined option
        # if ta options are discovered here, the ia_options value of this transaction instance will be set
        for option in options:
            if option in OPTIONS:
                OPTIONS[option].initialize(transaction=self, option=options[option])

    def get_options_string(self):
        """
        get all options in one string for debugging
        """
        options_string = ''
        # put own attributes into a string
        options = sorted(self.__dict__.keys())
        # options.sort()
        for option in options:
            # ignore some attributes
            if option not in IGNORED_LOG_OPTIONS and \
               not self.__dict__[option] in EMPTY_OPTIONS:
                if option == 'addresses':
                    if (CONST.OPTION.IA_NA or CONST.OPTION.IA_TA) in self.ia_options:
                        option_string = f'{option}:'
                        for address in self.__dict__[option]:
                            option_string += f' {colonify_ip6(address)}'
                        options_string = f'{options_string} | {option_string}'
                elif option == 'prefixes':
                    if CONST.OPTION.IA_PD in self.ia_options:
                        option_string = f'{option}:'
                        for p in self.__dict__[option]:
                            prefix, length = split_prefix(p)
                            option_string += combine_prefix_length(colonify_ip6(prefix), length)
                elif option == 'client_llip':
                    option_string = f'{option}: {colonify_ip6(self.__dict__[option])}'
                    options_string = f'{options_string} | {option_string}'
                elif option == 'mac':
                    if self.__dict__[option] != DUMMY_MAC:
                        # option_string = f'{option}: {str(self.__dict__[option])}'
                        option_string = f'{option}: {self.__dict__[option]}'
                        options_string = f'{options_string} | {option_string}'
                else:
                    # option_string = f'{option}: {str(self.__dict__[option])}'
                    option_string = f'{option}: {self.__dict__[option]}'
                    options_string = f'{options_string} | {option_string}'

        return options_string



#!/usr/bin/env python3

# dhcpy6d - DHCPv6 server
# Copyright (C) 2012-2022 Henri Wahl <h.wahl@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the
#
# Free Software Foundation
# 51 Franklin Street, Fifth Floor
# Boston, MA 02110-1301
# USA

import os
import os.path
from setuptools import setup, find_packages
import shutil

# workaround to get dhcpy6d-startscript created
try:
    if not os.path.exists('sbin'):
        os.mkdir('sbin')
    shutil.copyfile('main.py', 'sbin/dhcpy6d')
    os.chmod('sbin/dhcpy6d', 0o554)
except:
    print('could not copy main.py to sbin/dhcpy6d')

classifiers = [
    'Intended Audience :: System Administrators',
    'Development Status :: 5 - Production/Stable',
    'License :: OSI Approved :: GNU General Public License (GPL)',
    'Operating System :: POSIX :: Linux',
    'Operating System :: POSIX',
    'Natural Language :: English',
    'Programming Language :: Python',
    'Topic :: System :: Networking'
]

data_files = [('/var/lib/dhcpy6d', ['var/lib/volatile.sqlite']),
              ('/var/log', ['var/log/dhcpy6d.log']),
              ('/usr/share/doc/dhcpy6d', ['doc/clients-example.conf',
                                          'doc/config.sql',
                                          'doc/dhcpy6d-example.conf',
                                          'doc/dhcpy6d-minimal.conf',
                                          'doc/LICENSE',
                                          'doc/volatile.sql',
                                          'doc/volatile.postgresql']),
              ('/usr/share/man/man5', ['man/man5/dhcpy6d.conf.5',
                                       'man/man5/dhcpy6d-clients.conf.5']),
              ('/usr/share/man/man8', ['man/man8/dhcpy6d.8']),
              ('/etc', ['etc/dhcpy6d.conf']),
              ('/usr/sbin', ['sbin/dhcpy6d']),
              ]

setup(name='dhcpy6d',
      version='1.2.2',
      license='GNU GPL v2',
      description='DHCPv6 server daemon',
      long_description='Dhcpy6d delivers IPv6 addresses for DHCPv6 clients, which can be identified by DUID, hostname or MAC address as in the good old IPv4 days. It allows easy dualstack transition, addresses may be generated randomly, by range, by DNS, by arbitrary ID or MAC address. Clients can get more than one address, leases and client configuration can be stored in databases and DNS can be updated dynamically.',
      author='Henri Wahl',
      author_email='henri@dhcpy6d.de',
      url='https://dhcpy6d.de/',
      download_url='https://dhcpy6d.de/download',
      requires=['distro', 'dnspython'],
      packages=find_packages(),
      classifiers=classifiers,
      data_files=data_files
      )



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""Module dhcpy6d"""

from binascii import hexlify
from random import randint
import select
import shlex
import socket
import struct
import subprocess
import sys
import traceback

from .config import cfg
from .constants import (RTM_DELNEIGH,
                        RTM_GETNEIGH,
                        RTM_NEWNEIGH,
                        NLMSG_DONE,
                        NLMSG_ERROR,
                        MSG_HEADER_FLAGS,
                        MSG_HEADER_LENGTH,
                        MSG_HEADER_TYPE,
                        NUD_FAILED,
                        NUD_INCOMPLETE,
                        NUD_NOARP,
                        NDA,
                        NLA_ALIGNTO,
                        NLMSG_ALIGNTO)
from .globals import (collected_macs,
                      IF_NUMBER,
                      NC,
                      OS,
                      timer)
from .helpers import (colonify_ip6,
                      colonify_mac,
                      correct_mac,
                      decompress_ip6,
                      NeighborCacheRecord)
from .log import log
from .storage import volatile_store


def get_neighbor_cache_linux(if_number, now):
    """
    imported version of https://github.com/vokac/dhcpy6d
    https://github.com/vokac/dhcpy6d/commit/bd34d3efb18ba6016a2b3afea0b6a3fcdfb524a4
    Thanks for donating!
    """
    # open raw NETLINK socket
    # NETLINK_ROUTE has neighbor cache information too
    s = socket.socket(socket.AF_NETLINK, socket.SOCK_RAW, socket.NETLINK_ROUTE)
    # PID 0 means AUTOPID, let socket choose
    s.bind((0, 0))
    pid, groups = s.getsockname()

    # random sequence for NETLINK access
    seq = randint(0, pow(2, 31))

    # netlink message header (struct nlmsghdr)
    msg_header = struct.pack('IHHII',
                             MSG_HEADER_LENGTH,
                             MSG_HEADER_TYPE,
                             MSG_HEADER_FLAGS,
                             seq,
                             pid)

    # NETLINK message is always the same except header seq (struct ndmsg)
    msg = struct.pack('B', socket.AF_INET6)

    # send message with header
    s.send(msg_header + msg)

    # read all data from socket
    answer = b''
    while True:
        r, w, e = select.select([s], [], [], 0.)
        if s not in r:
            break  # no more data
        answer += s.recv(16384)

    result = {}
    curr_pos = 0
    answer_pos = 0
    answer_len = len(answer)

    nlmsghdr_fmt = 'IHHII'  # struct nlmsghdr
    nlattr_fmt = 'HH'  # struct nlattr
    ndmsg_fmt = 'BBHiHBB'  # struct ndmsg

    nlmsg_header_len = (struct.calcsize(nlmsghdr_fmt) + NLMSG_ALIGNTO - 1) & ~(NLMSG_ALIGNTO - 1)  # alignment to 4
    nla_header_len = (struct.calcsize(nlattr_fmt) + NLA_ALIGNTO - 1) & ~(NLA_ALIGNTO - 1)  # alignment to 4

    # parse netlink answer to RTM_GETNEIGH
    try:
        while answer_pos < answer_len:
            curr_pos = answer_pos

            nlmsg_len, nlmsg_type, nlmsg_flags, nlmsg_seq, nlmsg_pid = struct.unpack_from(f'<{nlmsghdr_fmt}',
                                                                                          answer,
                                                                                          answer_pos)
            # basic safety checks for received data (imitates NLMSG_OK)
            if nlmsg_len < struct.calcsize(f'<{nlmsghdr_fmt}'):
                log.warn('broken data from netlink (position {0}, nlmsg_len {1}): '
                         'nlmsg_len is smaller than structure size'.format(answer_pos, nlmsg_len))
                break
            if answer_len - answer_pos < struct.calcsize(f'<{nlmsghdr_fmt}'):
                log.warn(f'broken data from netlink (position {answer_pos}, length avail {answer_len - answer_pos}): '
                         'received data size is smaller than structure size')
                break
            if answer_len - answer_pos < nlmsg_len:
                log.warn(f'broken data from netlink (position {answer_pos}, length avail {answer_len - answer_pos}): '
                         'received dcolonify_ata size is smaller than nlmsg_len')
                break
            if pid != nlmsg_pid or seq != nlmsg_seq:
                log.warn(f'broken data from netlink (position {answer_pos}, length avail {answer_len - answer_pos}): '
                         f'invalid seq ({seq} x {nlmsg_seq}) or pid ({pid} x {nlmsg_pid})')
                break

            # data for this Routing/device hook record
            nlmsg_data = answer[answer_pos + nlmsg_header_len:answer_pos + nlmsg_len]

            if nlmsg_type == NLMSG_DONE:
                break
            if nlmsg_type == NLMSG_ERROR:
                nlmsgerr_error, nlmsgerr_len, nlmsgerr_type, nlmsgerr_flags, nlmsgerr_seq, nlmsgerr_pid = \
                    struct.unpack_from('<sIHHII', nlmsg_data)
                log.warn(f'broken data from netlink (position {answer_pos}, length avail {answer_len - answer_pos}): '
                         f'invalid message (errno {nlmsgerr_error})')
                break
            if nlmsg_type not in [RTM_NEWNEIGH, RTM_DELNEIGH, RTM_GETNEIGH]:
                log.warn(f'broken data from netlink (position {answer_pos}, length avail {answer_len - answer_pos}): '
                         f'this is really weird, wrong message type {nlmsg_type}')
                break

            curr_pos = answer_pos + nlmsg_header_len
            ndm_family, ndm_pad1, ndm_pad2, ndm_ifindex, ndm_state, ndm_flags, ndm_type = \
                struct.unpack_from(f'<{ndmsg_fmt}', nlmsg_data, 0)
            nda = {'NDM_FAMILY': ndm_family, 'NDM_IFINDEX': ndm_ifindex,
                   'NDM_STATE': ndm_state, 'NDM_FLAGS': ndm_flags,
                   'NDM_TYPE': ndm_type}
            nlmsg_data_pos = 0
            nlmsg_data_len = nlmsg_len - nlmsg_header_len
            while nlmsg_data_pos < nlmsg_data_len:
                curr_pos = answer_pos + nlmsg_header_len + nlmsg_data_pos

                nla_len, nla_type = \
                    struct.unpack_from(f'<{nlattr_fmt}', nlmsg_data, nlmsg_data_pos)

                # basic safety checks for received data (imitates RTA_OK)
                if nla_len < struct.calcsize(f'<{nlattr_fmt}'):
                    log.debug('This is normal for last record, but we should not get here (because of NLMSG_DONE); '
                              f'data size: {answer_len}, data[{answer_pos + nlmsg_header_len}:'
                              f'{answer_pos + nlmsg_len}] =  {hexlify(nlmsg_data)}')
                    break

                # data for this Routing/device hook record attribute
                nla_data = nlmsg_data[nlmsg_data_pos + nla_header_len:nlmsg_data_pos + nla_len]

                nda_type_key = NDA.get(nla_type, str(nla_type))
                if nda_type_key == 'NDA_DST':
                    nda[nda_type_key] = colonify_ip6(hexlify(nla_data))
                elif nda_type_key == 'NDA_LLADDR':
                    nda[nda_type_key] = colonify_mac(hexlify(nla_data))
                elif nda_type_key == 'NDA_CACHEINFO':
                    nda[nda_type_key] = struct.unpack_from('<IIII', nla_data)
                elif nda_type_key == 'NDA_VLAN':
                    nda[nda_type_key] = hexlify(nla_data)
                else:
                    nda[nda_type_key] = nla_data

                nlmsg_data_pos += nla_header_len
                nlmsg_data_pos += (nla_len - nla_header_len + NLA_ALIGNTO - 1) & ~(NLA_ALIGNTO - 1)  # alignment to 4

            # prepare all required data to be returned to callee
            # * only care about configured devices
            # * no need for multicast address cache entries (MAC 33:33:...)
            if nda['NDM_STATE'] & ~(NUD_INCOMPLETE | NUD_FAILED | NUD_NOARP):
                if nda['NDM_IFINDEX'] not in if_number:
                    log.debug(f"can't find device for interface index {nda['NDM_IFINDEX']}")
                elif 'NDA_DST' not in nda:
                    log.warn(f"can't find destination address (wrong entry state: {nda['NDM_STATE']}?!)")
                elif 'NDA_LLADDR' not in nda:
                    log.warn(f"can't find local hardware address (wrong entry state: {nda['NDM_STATE']}?!)")
                else:
                    if if_number[nda['NDM_IFINDEX']] in cfg.INTERFACE and not nda['NDA_LLADDR'].startswith('33:33:'):
                        # store neighbor caches entries
                        record = NeighborCacheRecord(llip=decompress_ip6(nda['NDA_DST']),
                                                     mac=nda['NDA_LLADDR'],
                                                     interface=if_number[nda['NDM_IFINDEX']],
                                                     now=now)
                        result[record.llip] = record

            # move to next record
            answer_pos += nlmsg_len

    except struct.error as e:
        log.warn(f'broken data from netlink (position {answer_pos}, '
                 f'data[{curr_pos}:{answer_len}] = {hexlify(answer[curr_pos:curr_pos + 8])}...): {str(e)}')

    # clean up
    s.close()

    return result


def collect_macs(now):
    """
    collect MAC address from clients to link local addresses with MACs
    if a client has a new MAC the LLIP changes - with privacy extension enabled anyway
    calls local ip command to get neighbor cache - any more sophisticated idea is welcome!

    The Linux netlink method is considered stable now.
    """
    try:
        # Linux can use kernel neighbor cache
        if OS == 'Linux':
            for host in list(get_neighbor_cache_linux(IF_NUMBER, timer.time).values()):
                if host.llip not in collected_macs:
                    if host.llip.startswith('fe80'):
                        collected_macs[host.llip] = host
                        if cfg.LOG_MAC_LLIP:
                            log.info(f'collected mac {host.mac} for llip {colonify_ip6(host.llip)}')
                        if cfg.CACHE_MAC_LLIP:
                            volatile_store.store_mac_llip(host.mac, host.llip, timer.time)
        else:
            # subject to change - other distros might have other paths - might become a task
            # for a setup routine to find appropriate paths
            for host in subprocess.getoutput(NC[OS]['call']).splitlines():
                # get fragments of output line
                frags = shlex.split(host)
                if frags[NC[OS]['dev']] in cfg.INTERFACE and len(frags) >= NC[OS]['len']:
                    # get rid of %interface
                    frags[NC[OS]['llip']] = decompress_ip6(frags[NC[OS]['llip']].split('%')[0])
                    if frags[NC[OS]['mac']] == '(incomplete)':
                        continue
                    # correct maybe shortened MAC
                    frags[NC[OS]['mac']] = correct_mac(frags[NC[OS]['mac']])
                    # put non yet existing LLIPs into dictionary - if they have MACs
                    if not frags[NC[OS]['llip']] in collected_macs and \
                       frags[NC[OS]['llip']].lower().startswith('fe80') and \
                       ':' in frags[NC[OS]['mac']]:
                        collected_macs[frags[NC[OS]['llip']]] = NeighborCacheRecord(llip=frags[NC[OS]['llip']],
                                                                                    mac=frags[NC[OS]['mac']],
                                                                                    interface=frags[NC[OS]['dev']],
                                                                                    now=now)
                        if cfg.LOG_MAC_LLIP:
                            log.info(f"collected mac {frags[NC[OS]['mac']]} for "
                                     f"llip {colonify_ip6(frags[NC[OS]['llip']])}")
                        volatile_store.store_mac_llip(frags[NC[OS]['mac']], frags[NC[OS]['llip']], timer.time)
    except Exception as err:
        traceback.print_exc(file=sys.stdout)
        sys.stdout.flush()
        log.error('collect_macs(): ' + str(err))



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import sys

from ..config import cfg

# put SQL schemas here to be in reach of all storage types
# generic mostly usable for SQLite and MySQL/MariaDB
MYSQL_SQLITE = {}
MYSQL_SQLITE['meta'] = '''
                    CREATE TABLE meta (
                    item_key varchar(255) NOT NULL,
                    item_value varchar(255) NOT NULL,
                    PRIMARY KEY (item_key)
                    );
                    '''
MYSQL_SQLITE['leases'] = '''
                    CREATE TABLE leases (
                    address varchar(32) NOT NULL,
                    active tinyint(4) NOT NULL,
                    preferred_lifetime int(11) NOT NULL,
                    valid_lifetime int(11) NOT NULL,
                    hostname varchar(255) NOT NULL,
                    type varchar(255) NOT NULL,
                    category varchar(255) NOT NULL,
                    ia_type varchar(255) NOT NULL,
                    class varchar(255) NOT NULL,
                    mac varchar(17) NOT NULL,
                    duid varchar(255) NOT NULL,
                    last_update bigint NOT NULL,
                    preferred_until bigint NOT NULL,
                    valid_until bigint NOT NULL,
                    iaid varchar(8) DEFAULT NULL,
                    last_message int(11) NOT NULL DEFAULT 0,
                    PRIMARY KEY (address)
                    );
                    '''
MYSQL_SQLITE['macs_llips'] = '''
                        CREATE TABLE macs_llips (
                        mac varchar(17) NOT NULL,
                        link_local_ip varchar(39) NOT NULL,
                        last_update bigint NOT NULL,
                        PRIMARY KEY (mac)
                        );
                        '''
MYSQL_SQLITE['prefixes'] = '''
                    CREATE TABLE prefixes (
                    prefix varchar(32) NOT NULL,
                    length tinyint(4) NOT NULL,
                    active tinyint(4) NOT NULL,
                    preferred_lifetime int(11) NOT NULL,
                    valid_lifetime int(11) NOT NULL,
                    hostname varchar(255) NOT NULL,
                    type varchar(255) NOT NULL,
                    category varchar(255) NOT NULL,
                    class varchar(255) NOT NULL,
                    mac varchar(17) NOT NULL,
                    duid varchar(255) NOT NULL,
                    last_update bigint NOT NULL,
                    preferred_until bigint NOT NULL,
                    valid_until bigint NOT NULL,
                    iaid varchar(8) DEFAULT NULL,
                    last_message int(11) NOT NULL DEFAULT 0,
                    PRIMARY KEY (prefix)
                    );
                    '''
MYSQL_SQLITE['routes'] = '''
                    CREATE TABLE routes (
                    prefix varchar(32) NOT NULL,
                    length tinyint(4) NOT NULL,
                    router varchar(32) NOT NULL,
                    last_update bigint NOT NULL,
                    PRIMARY KEY (prefix)
                    );
                    '''

# Postgresql has some differences and so its own schemas
POSTGRESQL_SCHEMA = {}
POSTGRESQL_SCHEMA['meta'] = '''
                        CREATE TABLE meta (
                        item_key varchar(255) NOT NULL,
                        item_value varchar(255) NOT NULL,
                        PRIMARY KEY (item_key)
                        );
                        
                        '''
POSTGRESQL_SCHEMA['leases'] = '''
                            CREATE TABLE leases (
                            address varchar(32) NOT NULL,
                            active smallint NOT NULL,
                            preferred_lifetime int NOT NULL,
                            valid_lifetime int NOT NULL,
                            hostname varchar(255) NOT NULL,
                            type varchar(255) NOT NULL,
                            category varchar(255) NOT NULL,
                            ia_type varchar(255) NOT NULL,
                            class varchar(255) NOT NULL,
                            mac varchar(17) NOT NULL,
                            duid varchar(255) NOT NULL,
                            last_update bigint NOT NULL,
                            preferred_until bigint NOT NULL,
                            valid_until bigint NOT NULL,
                            iaid varchar(8) DEFAULT NULL,
                            last_message int NOT NULL DEFAULT 0,
                            PRIMARY KEY (address)
                            );
                            '''
POSTGRESQL_SCHEMA['macs_llips'] = '''
                                CREATE TABLE macs_llips (
                                mac varchar(17) NOT NULL,
                                link_local_ip varchar(39) NOT NULL,
                                last_update bigint NOT NULL,
                                PRIMARY KEY (mac)
                                );
                                '''
POSTGRESQL_SCHEMA['prefixes'] = '''
                            CREATE TABLE prefixes (
                            prefix varchar(32) NOT NULL,
                            length smallint NOT NULL,
                            active smallint NOT NULL,
                            preferred_lifetime int NOT NULL,
                            valid_lifetime int NOT NULL,
                            hostname varchar(255) NOT NULL,
                            type varchar(255) NOT NULL,
                            category varchar(255) NOT NULL,
                            class varchar(255) NOT NULL,
                            mac varchar(17) NOT NULL,
                            duid varchar(255) NOT NULL,
                            last_update bigint NOT NULL,
                            preferred_until bigint NOT NULL,
                            valid_until bigint NOT NULL,
                            iaid varchar(8) DEFAULT NULL,
                            last_message int NOT NULL DEFAULT 0,
                            PRIMARY KEY (prefix)
                            );
                            '''
POSTGRESQL_SCHEMA['routes'] = '''
                            CREATE TABLE routes (
                            prefix varchar(32) NOT NULL,
                            length smallint NOT NULL,
                            router varchar(32) NOT NULL,
                            last_update bigint NOT NULL,
                            PRIMARY KEY (prefix)
                            );
                            '''


# formerly part of store.py but might fit better here
def legacy_adjustments(db):
    """
    adjust some existing data to work with newer versions of dhcpy6d
    """
    try:
        if db.query('SELECT last_message FROM leases LIMIT 1') is None:
            # row 'last_message' in schemas 'leases' does not exist yet, comes with version 0.1.6
            db.query('ALTER TABLE leases ADD last_message INT NOT NULL DEFAULT 0')
            print("Adding row 'last_message' to table 'leases' in volatile storage succeeded.")
    except:
        print("\n'ALTER TABLE leases ADD last_message INT NOT NULL DEFAULT 0' on volatile database failed.")
        print('Please apply manually or grant necessary permissions.\n')
        sys.exit(1)

    # after 0.4.3 with working PostgreSQL support the timestamps have to be stores in epoch seconds, not datetime
    # also after 0.4.3 there will be a third table containing meta information - for a first start it should contain
    # a database version number
    try:
        try:
            # only newer databases contain a version number - real ones starting with 1
            # non-existing version is False
            db_version = db.get_db_version()
            if db_version is None or int(db_version) == 0:
                db_operations = []
                # only create meta table if it does not exist yet
                if not 'meta' in db.get_tables():
                    db_operations.append('CREATE TABLE meta (item_key varchar(255) NOT NULL,\
                                  item_value varchar(255) NOT NULL, PRIMARY KEY (item_key))')
                # add version of database scheme
                db_operations.append("INSERT INTO meta (item_key, item_value) VALUES ('version', '1')")
                for db_operation in db_operations:
                    db.query(db_operation)
                    print(f"{db_operation} in volatile storage succeded.")
        except Exception as err:
            print(f"\n{db_operation} on volatile database failed.")
            print('Please apply manually or grant necessary permissions.\n')
            sys.exit(1)
    except Exception as err:
        print('\nSomething went wrong when retrieving version from database.\n')
        sys.exit(1)

    # find out if timestamps still are in datetime format - applies only to sqlite and mysql anyway
    if cfg.STORE_VOLATILE in ['sqlite', 'mysql']:
        db_datetime_test = db.query('SELECT last_update FROM leases LIMIT 1')
        if len(db_datetime_test) > 0:
            import datetime

            # flag to find out which update has to be done
            update_type = False

            # MySQL
            if type(db_datetime_test[0][0]) is datetime.datetime:
                update_type = 'mysql'

            # SQLite
            if type(db_datetime_test[0][0]) is str:
                if ' ' in db_datetime_test[0][0]:
                    update_type = 'sqlite'

            if update_type:
                # add new columns with suffix *_new
                db_tables = {'leases': ['last_update', 'preferred_until', 'valid_until'],
                             'macs_llips': ['last_update']}

                if update_type == 'mysql':
                    for table in db_tables:
                        for column in db_tables[table]:
                            db.query(f'ALTER TABLE {table} ADD COLUMN {column}_new bigint NOT NULL')
                            print(f'ALTER TABLE {table} ADD COLUMN {column}_new bigint NOT NULL succeeded')
                    # get old timestamps
                    timestamps_old = db.query(
                        'SELECT address, last_update, preferred_until, valid_until FROM leases')
                    for timestamp_old in timestamps_old:
                        address, last_update, preferred_until, valid_until = timestamp_old
                        # convert SQLite datetime values from unicode to Python datetime
                        if update_type == 'sqlite':
                            last_update = datetime.datetime.strptime(last_update, '%Y-%m-%d %H:%M:%S.%f')
                            preferred_until = datetime.datetime.strptime(preferred_until, '%Y-%m-%d %H:%M:%S.%f')
                            valid_until = datetime.datetime.strptime(valid_until, '%Y-%m-%d %H:%M:%S.%f')

                        last_update_new = last_update.strftime('%s')
                        preferred_until_new = preferred_until.strftime('%s')
                        valid_until_new = valid_until.strftime('%s')
                        db.query(f"UPDATE leases SET last_update_new = {last_update_new}, "
                                 f"preferred_until_new = {preferred_until_new}, "
                                 f"valid_until_new = {valid_until_new} "
                                 f"WHERE address = '{address}'")
                    print('Converting timestamps of leases succeeded')
                    timestamps_old = db.query('SELECT mac, last_update FROM macs_llips')
                    for timestamp_old in timestamps_old:
                        mac, last_update = timestamp_old
                        last_update_new = last_update.strftime('%s')
                        db.query(f"UPDATE macs_llips SET last_update_new = {last_update_new} WHERE mac = '{mac}'")
                    print('Converting timestamps of macs_llips succeeded')
                    for table in db_tables:
                        for column in db_tables[table]:
                            db.query(f'ALTER TABLE {table} DROP COLUMN {column}')
                            db.query(f'ALTER TABLE {table} CHANGE COLUMN {column}_new {column} BIGINT NOT NULL')
                            print(f'Moving column {column} of table {table} succeeded')

                if update_type == 'sqlite':
                    for table in db_tables:
                        db.query(f'ALTER TABLE {table} RENAME TO {table}_old')

                    db.query('CREATE TABLE leases AS SELECT address,active,last_message,preferred_lifetime,'
                             'valid_lifetime,hostname,type,category,ia_type,'
                             'class,mac,duid,iaid '
                             'FROM leases_old')

                    db.query('CREATE TABLE macs_llips AS SELECT mac,link_local_ip FROM macs_llips_old')

                    # add timestamp columns in bigint format instead of datetime
                    for table in db_tables:
                        for column in db_tables[table]:
                            db.query(f'ALTER TABLE {table} ADD COLUMN {column} bigint')

                    # get old timestamps
                    timestamps_old = db.query(
                        'SELECT address, last_update, preferred_until, valid_until FROM leases_old')
                    for timestamp_old in timestamps_old:
                        address, last_update, preferred_until, valid_until = timestamp_old
                        # convert SQLite datetime values from unicode to Python datetime
                        if update_type == 'sqlite':
                            last_update = datetime.datetime.strptime(last_update, '%Y-%m-%d %H:%M:%S.%f')
                            preferred_until = datetime.datetime.strptime(preferred_until, '%Y-%m-%d %H:%M:%S.%f')
                            valid_until = datetime.datetime.strptime(valid_until, '%Y-%m-%d %H:%M:%S.%f')

                        last_update_new = last_update.strftime('%s')
                        preferred_until_new = preferred_until.strftime('%s')
                        valid_until_new = valid_until.strftime('%s')
                        db.query(f"UPDATE leases SET last_update = {last_update_new}, "
                                 f"preferred_until = {preferred_until_new}, "
                                 f"valid_until = {valid_until_new} "
                                 f"WHERE address = '{address}'")
                    print('Converting timestamps of leases succeeded')
                    timestamps_old = db.query('SELECT mac, last_update FROM macs_llips_old')
                    for timestamp_old in timestamps_old:
                        mac, last_update = timestamp_old
                        last_update_new = last_update.strftime('%s')
                        db.query(f"UPDATE macs_llips SET last_update = {last_update_new} WHERE mac = '{mac}'")
                    print('Converting timestamps of macs_llips succeeded')

    # Extend volatile database to handle prefixes - comes with database version 2
    if int(db.get_db_version()) < 2:
        if not 'prefixes' in db.get_tables():
            if cfg.STORE_VOLATILE in ['sqlite', 'mysql']:
                db.query('CREATE TABLE prefixes (\
                              prefix varchar(32) NOT NULL,\
                              length tinyint(4) NOT NULL,\
                              active tinyint(4) NOT NULL,\
                              preferred_lifetime int(11) NOT NULL,\
                              valid_lifetime int(11) NOT NULL,\
                              hostname varchar(255) NOT NULL,\
                              type varchar(255) NOT NULL,\
                              category varchar(255) NOT NULL,\
                              class varchar(255) NOT NULL,\
                              mac varchar(17) NOT NULL,\
                              duid varchar(255) NOT NULL,\
                              last_update bigint NOT NULL,\
                              preferred_until bigint NOT NULL,\
                              valid_until bigint NOT NULL,\
                              iaid varchar(8) DEFAULT NULL,\
                              last_message int(11) NOT NULL DEFAULT 0,\
                              PRIMARY KEY (prefix)\
                            )')

            elif cfg.STORE_VOLATILE == 'postgresql':
                db.query('CREATE TABLE prefixes (\
                              prefix varchar(32) NOT NULL,\
                              length smallint NOT NULL,\
                              active smallint NOT NULL,\
                              preferred_lifetime int NOT NULL,\
                              valid_lifetime int NOT NULL,\
                              hostname varchar(255) NOT NULL,\
                              type varchar(255) NOT NULL,\
                              category varchar(255) NOT NULL,\
                              class varchar(255) NOT NULL,\
                              mac varchar(17) NOT NULL,\
                              duid varchar(255) NOT NULL,\
                              last_update bigint NOT NULL,\
                              preferred_until bigint NOT NULL,\
                              valid_until bigint NOT NULL,\
                              iaid varchar(8) DEFAULT NULL,\
                              last_message int NOT NULL DEFAULT 0,\
                              PRIMARY KEY (prefix)\
                            )')

        # increase version to 2
        db.query("UPDATE meta SET item_value='2' WHERE item_key='version'")

        # All OK
        print("Table 'prefixes' is OK")

    # Extend volatile database to handle routes - comes with database version 3
    if int(db.get_db_version()) < 3:
        if not 'prefixes' in db.get_tables():
            if cfg.STORE_VOLATILE in ['sqlite', 'mysql']:
                db.query('CREATE TABLE routes (\
                              prefix varchar(32) NOT NULL,\
                              length tinyint(4) NOT NULL,\
                              router varchar(32) NOT NULL,\
                              last_update bigint NOT NULL,\
                              PRIMARY KEY (prefix)\
                            )')

            elif cfg.STORE_VOLATILE == 'postgresql':
                db.query('CREATE TABLE routes (\
                              prefix varchar(32) NOT NULL,\
                              length smallint NOT NULL,\
                              router varchar(32) NOT NULL,\
                              last_update bigint NOT NULL,\
                              PRIMARY KEY (prefix)\
                            )')

        # increase version to 3
        db.query("UPDATE meta SET item_value='3' WHERE item_key='version'")

        # All OK
        print("Table 'routes' is OK")



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import sys
import threading
import traceback

from ..config import cfg
from ..globals import (config_answer_queue,
                       config_query_queue,
                       config_store,
                       volatile_answer_queue,
                       volatile_query_queue,
                       volatile_store)
from ..helpers import error_exit

from .mysql import DBMySQL
from .postgresql import DBPostgreSQL
from .sqlite import SQLite
from .store import (ClientConfig,
                    ClientConfigDicts,
                    Store)
from .textfile import Textfile


class QueryQueue(threading.Thread):
    """
    Pump queries around
    """
    def __init__(self, name='', store_type=None, query_queue=None, answer_queue=None):
        threading.Thread.__init__(self, name=name)
        self.query_queue = query_queue
        self.answer_queue = answer_queue
        self.store = store_type
        self.setDaemon(True)

    def run(self):
        """
        receive queries and ask the DB interface for answers which will be put into
        answer queue
        """
        while True:
            query = self.query_queue.get()
            try:
                answer = self.store.db_query(query)
            except Exception as error:
                traceback.print_exc(file=sys.stdout)
                sys.stdout.flush()
                answer = error

            self.answer_queue.put({query: answer})


# because of thread trouble there should not be too much db connections at once
# so we need to use the queryqueue way - subject to change
# source of configuration of hosts
# use client configuration only if needed
if cfg.STORE_CONFIG:
    if cfg.STORE_CONFIG == 'file':
        config_store = Textfile(config_query_queue, config_answer_queue)
    if cfg.STORE_CONFIG == 'mysql':
        config_store = DBMySQL(config_query_queue, config_answer_queue)
    if cfg.STORE_CONFIG == 'postgresql':
        config_store = DBPostgreSQL(config_query_queue, config_answer_queue)
    if cfg.STORE_CONFIG == 'sqlite':
        config_store = SQLite(config_query_queue, config_answer_queue, storage_type='config')
else:
    # dummy configstore if no client config is needed
    config_store = Store(config_query_queue, config_answer_queue)
    # 'none' store is always connected
    config_store.connected = True

# storage for changing data like leases, LLIPs, DUIDs etc.
if cfg.STORE_VOLATILE == 'mysql':
    volatile_store = DBMySQL(volatile_query_queue, volatile_answer_queue)
if cfg.STORE_VOLATILE == 'postgresql':
    volatile_store = DBPostgreSQL(volatile_query_queue, volatile_answer_queue)
if cfg.STORE_VOLATILE == 'sqlite':
    volatile_store = SQLite(volatile_query_queue, volatile_answer_queue, storage_type='volatile')

# do not start if no database connection exists
if not config_store.connected:
    error_exit('Configuration database is not connected!')
if not volatile_store.connected:
    error_exit('Database for volatile data is not connected!')



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

from ..config import (Address,
                      cfg,
                      Prefix)
from ..constants import CONST
from ..domain import get_ip_from_dns

from .parse_pattern import (parse_pattern_address,
                            parse_pattern_prefix)


def default(client=None, client_config=None, transaction=None):
    # use default class if host is unknown
    client.hostname = transaction.hostname
    client.client_class = 'default_' + transaction.interface
    # apply answer type of client to transaction - useful if no answer or no address available is configured
    transaction.answer = cfg.CLASSES[client.client_class].ANSWER

    if 'addresses' in cfg.CLASSES['default_' + transaction.interface].ADVERTISE and \
            (3 or 4) in transaction.ia_options:
        for address in cfg.CLASSES['default_' + transaction.interface].ADDRESSES:
            # addresses of category 'dns' will be searched in DNS
            if cfg.ADDRESSES[address].CATEGORY == 'dns':
                a = get_ip_from_dns(client.hostname)
            else:
                a = parse_pattern_address(cfg.ADDRESSES[address], client, transaction)
            if a:
                ia = Address(address=a, ia_type=cfg.ADDRESSES[address].IA_TYPE,
                             preferred_lifetime=cfg.ADDRESSES[address].PREFERRED_LIFETIME,
                             valid_lifetime=cfg.ADDRESSES[address].VALID_LIFETIME,
                             category=cfg.ADDRESSES[address].CATEGORY,
                             aclass=cfg.ADDRESSES[address].CLASS,
                             atype=cfg.ADDRESSES[address].TYPE,
                             dns_update=cfg.ADDRESSES[address].DNS_UPDATE,
                             dns_zone=cfg.ADDRESSES[address].DNS_ZONE,
                             dns_rev_zone=cfg.ADDRESSES[address].DNS_REV_ZONE,
                             dns_ttl=cfg.ADDRESSES[address].DNS_TTL)
                client.addresses.append(ia)

    if 'prefixes' in cfg.CLASSES['default_' + transaction.interface].ADVERTISE and \
            CONST.OPTION.IA_PD in transaction.ia_options:

        for prefix in cfg.CLASSES['default_' + transaction.interface].PREFIXES:
            p = parse_pattern_prefix(cfg.PREFIXES[prefix], client_config, transaction)
            # in case range has been exceeded p will be None
            if p:
                ia_pd = Prefix(prefix=p,
                               length=cfg.PREFIXES[prefix].LENGTH,
                               preferred_lifetime=cfg.PREFIXES[prefix].PREFERRED_LIFETIME,
                               valid_lifetime=cfg.PREFIXES[prefix].VALID_LIFETIME,
                               category=cfg.PREFIXES[prefix].CATEGORY,
                               pclass=cfg.PREFIXES[prefix].CLASS,
                               ptype=cfg.PREFIXES[prefix].TYPE,
                               route_link_local=cfg.PREFIXES[prefix].ROUTE_LINK_LOCAL)
                client.prefixes.append(ia_pd)

    # given client has been modified successfully
    return True
